"""Mesoscope Extractor class"""

import argparse
import json
import logging
import sys
from pathlib import Path
from typing import Tuple, Union

import h5py as h5
import tifffile

from aind_metadata_extractor.core import BaseExtractor
from aind_metadata_extractor.mesoscope.job_settings import JobSettings
from aind_metadata_extractor.utils.camstim_sync.camstim import Camstim, CamstimSettings

from aind_metadata_extractor.models.mesoscope import MesoscopeExtractModel


class MesoscopeExtract(BaseExtractor):
    """Class to manage transforming mesoscope platform json and metadata into
    a mesoscope model model."""

    _STRUCTURE_LOOKUP_DICT = {
        385: "VISp",
        394: "VISam",
        402: "VISal",
        409: "VISl",
        417: "VISrl",
        533: "VISpm",
        312782574: "VISli",
    }

    # TODO: Deprecate this constructor. Use GenericEtl constructor instead
    def __init__(self, job_settings: Union[JobSettings, str]):
        """
        Class constructor for MesoscopeExtract.
        Parameters
        ----------
        job_settings: Union[JobSettings, str]
          Variables for a particular session
        """
        if isinstance(job_settings, str):
            job_settings_model = JobSettings.model_validate_json(job_settings)
        else:
            job_settings_model = job_settings
        if isinstance(job_settings_model.behavior_source, str):
            job_settings_model.behavior_source = Path(job_settings_model.behavior_source)
        camstim_output = job_settings_model.output_directory
        if job_settings_model.make_camsitm_dir:
            camstim_output = job_settings_model.output_directory / f"{job_settings_model.session_id}_behavior"
        self.job_settings = job_settings_model
        camstim_settings = CamstimSettings(
            input_source=self.job_settings.behavior_source,
            output_directory=camstim_output,
            session_id=self.job_settings.session_id,
            subject_id=self.job_settings.subject_id,
        )
        self.camstim = Camstim(camstim_settings)

    @staticmethod
    def _read_metadata(tiff_path: Path):
        """
        Calls tifffile.read_scanimage_metadata on the specified
        path and returns teh result. This method was factored
        out so that it could be easily mocked in unit tests.
        """

        with open(tiff_path, "rb") as tiff:
            file_handle = tifffile.FileHandle(tiff)
            file_contents = tifffile.read_scanimage_metadata(file_handle)
        return file_contents

    def _read_h5_metadata(self, h5_path: str):
        """Reads scanimage metadata from h5path

        Parameters
        ----------
        h5_path : str
            Path to h5 file

        Returns
        -------
        dict
        """
        data = h5.File(h5_path)
        try:
            file_contents = data["scanimage_metadata"][()].decode()
        except KeyError:
            logging.warning("No scanimage metadata found in h5 file. Returning image shape 512x512.")
            file_contents = '[{"SI.hRoiManager.pixelsPerLine": 512, "SI.hRoiManager.linesPerFrame": 512}]'  # noqa
        data.close()
        file_contents = json.loads(file_contents)
        return file_contents

    def _extract_behavior_metdata(self) -> dict:
        """Loads behavior metadata from the behavior json files
        Returns
        -------
        dict
            behavior video metadata
        """
        session_metadata = {}
        session_id = self.job_settings.session_id
        for ftype in sorted(list(self.job_settings.behavior_source.glob("*json"))):
            if (
                ("Behavior" in ftype.stem and session_id in ftype.stem)
                or ("Eye" in ftype.stem and session_id in ftype.stem)
                or ("Face" in ftype.stem and session_id in ftype.stem)
            ):
                with open(ftype, "r") as f:
                    session_metadata[ftype.stem] = json.load(f)
        return session_metadata

    def _extract_platform_metadata(self, session_metadata: dict) -> dict:
        """Parses the platform json file and returns the metadata

        Parameters
        ----------
        session_metadata : dict
            For session parsing

        Returns
        -------
        dict
            _description_
        """
        input_source = next(self.job_settings.input_source.glob("*platform.json"), "")
        if (isinstance(input_source, str) and input_source == "") or not input_source.exists():
            raise ValueError("No platform json file found in directory")
        with open(input_source, "r") as f:
            session_metadata["platform"] = json.load(f)

        return session_metadata

    def _extract_time_series_metadata(self) -> dict:
        """Grab time series metadata from TIFF or HDF5

        Returns
        -------
        dict
            timeseries metadata
        """
        timeseries = next(self.job_settings.input_source.glob("*timeseries*.tiff"), "")
        if timeseries:
            meta = self._read_metadata(timeseries)
        else:
            experiment_dir = list(self.job_settings.input_source.glob("ophys_experiment*"))[0]
            experiment_id = experiment_dir.name.split("_")[-1]
            timeseries = next(experiment_dir.glob(f"{experiment_id}.h5"))
            meta = self._read_h5_metadata(str(timeseries))

        return meta

    def _extract(self) -> MesoscopeExtractModel:
        """extract data from the platform json file and tiff file (in the
        future).
        If input source is a file, will extract the data from the file.
        The input source is a directory, will extract the data from the
        directory.

        Returns
        -------
        (dict, dict)
            The extracted data from the platform json file and the time series
        """
        # The pydantic models will validate that the user inputs a Path.
        # We can add validators there if we want to coerce strings to Paths.
        session_metadata = self._extract_behavior_metdata()
        session_metadata = self._extract_platform_metadata(session_metadata)
        meta = self._extract_time_series_metadata()
        epochs, session_type = self._camstim_epoch_and_session()
        user_settings = self.job_settings.model_dump()
        data = {
            "session_metadata": session_metadata,
            "camstim_epochs": epochs,
            "camstim_session_type": session_type,
            "time_series_header": meta,
            "job_settings": user_settings,
        }
        return MesoscopeExtractModel(
            tiff_header=data["time_series_header"],
            session_metadata=data["session_metadata"],
            camstim_epchs=data["camstim_epochs"],
            camstim_session_type=data["camstim_session_type"],
            job_settings=data["job_settings"],
        )

    def _camstim_epoch_and_session(self) -> Tuple[list, str]:
        """Get the camstim table and epochs

        Returnsd
        -------
        list
            The camstim table and epochs
        """
        if self.camstim.behavior:
            self.camstim.build_behavior_table()
        else:
            self.camstim.build_stimulus_table(modality="ophys")
        return self.camstim.epochs_from_stim_table(), self.camstim.session_type

    def run_job(self) -> dict:
        """
        Run the extraction job.
        """
        self.metadata = self._extract()
        return self.metadata.model_dump()

    @classmethod
    def from_args(cls, args: list):
        """
        Adds ability to construct settings from a list of arguments.
        Parameters
        ----------
        args : list
        A list of command line arguments to parse.
        """

        logging.warning("This method will be removed in future versions. " "Please use JobSettings.from_args instead.")

        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-u",
            "--job-settings",
            required=True,
            type=json.loads,
            help=("""
                Custom settings defined by the user defined as a json
                 string. For example: -u
                 '{"experimenter_full_name":["John Smith","Jane Smith"],
                 "subject_id":"12345",
                 "session_start_time":"2023-10-10T10:10:10",
                 "session_end_time":"2023-10-10T18:10:10",
                 "project":"my_project"}
                """),
        )
        job_args = parser.parse_args(args)
        job_settings_from_args = JobSettings(**job_args.job_settings)
        return cls(
            job_settings=job_settings_from_args,
        )


if __name__ == "__main__":
    sys_args = sys.argv[1:]
    main_job_settings = JobSettings.from_args(sys_args)
    metl = MesoscopeExtract(job_settings=main_job_settings)
    extracted_data = metl.run_job()



"""Module to extract metadata from tif files."""

import bisect
import json
import logging
import os
import re
import sys
from pathlib import Path
from typing import Dict, List

from ScanImageTiffReader import ScanImageTiffReader

from aind_metadata_extractor.bergamo.settings import Settings
from aind_metadata_extractor.core import BaseExtractor
from aind_metadata_extractor.models.bergamo import ExtractedInfo, ExtractedInfoItem, RawImageInfo, TifFileGroup


class Extractor(BaseExtractor):
    """Class to manage extracting metadata from files."""

    def __init__(self, settings: Settings):
        """Class constructor."""

        self.settings = settings

    def get_tif_file_locations(self) -> Dict[str, List[Path]]:
        """Scans the input source directory and returns a dictionary of file
        groups in an ordered list. For example, if the directory had
        [neuron2_00001.tif, neuron2_00002.tif, stackPost_00001.tif,
        stackPost_00002.tif, stackPost_00003.tif], then it will return
        { "neuron2": [neuron2_00001.tif, neuron2_00002.tif],
         "stackPost":
           [stackPost_00001.tif, stackPost_00002.tif, stackPost_00003.tif]
        }
        """
        compiled_regex = re.compile(r"^(.*)_.*?(\d+).tif+$")
        tif_file_map = {}
        for root, dirs, files in os.walk(self.settings.input_source):
            for name in files:
                matched = re.match(compiled_regex, name)
                if matched:
                    groups = matched.groups()
                    file_stem = groups[0]
                    # tif_number = groups[1]
                    tif_filepath = Path(os.path.join(root, name))
                    if tif_file_map.get(file_stem) is None:
                        tif_file_map[file_stem] = [tif_filepath]
                    else:
                        bisect.insort(tif_file_map[file_stem], tif_filepath)

            # Only scan the top level files
            break
        return tif_file_map

    @staticmethod
    def flat_dict_to_nested(flat: dict, key_delim: str = ".") -> dict:
        """
        Utility method to convert a flat dictionary into a nested dictionary.
        Modified from https://stackoverflow.com/a/50607551
        Parameters
        ----------
        flat : dict
          Example {"a.b.c": 1, "a.b.d": 2, "e.f": 3}
        key_delim : str
          Delimiter on dictionary keys. Default is '.'.

        Returns
        -------
        dict
          A nested dictionary like {"a": {"b": {"c":1, "d":2}, "e": {"f":3}}
        """

        def __nest_dict_rec(k, v, out) -> None:
            """Simple recursive method being called."""
            k, *rest = k.split(key_delim, 1)
            if rest:
                __nest_dict_rec(rest[0], v, out.setdefault(k, {}))
            else:
                out[k] = v

        result = {}
        for flat_key, flat_val in flat.items():
            __nest_dict_rec(flat_key, flat_val, result)
        return result

    def extract_raw_info_from_file(self, file_path: Path) -> RawImageInfo:
        """
        Use ScanImageTiffReader to read metadata from a single file and parse
        it into a RawImageInfo object
        Parameters
        ----------
        file_path : Path

        Returns
        -------
        RawImageInfo

        """
        with ScanImageTiffReader(str(file_path)) as reader:
            reader_metadata = reader.metadata()
            reader_shape = reader.shape()
            reader_descriptions = [
                dict([(s.split(" = ", 1)[0], s.split(" = ", 1)[1]) for s in reader.description(i).strip().split("\n")])
                for i in range(0, len(reader))
            ]

        metadata_first_part = reader_metadata.split("\n\n")[0]
        flat_metadata_header_dict = dict(
            [(s.split(" = ", 1)[0], s.split(" = ", 1)[1]) for s in metadata_first_part.split("\n")]
        )
        metadata_dict = self.flat_dict_to_nested(flat_metadata_header_dict)
        reader_metadata_json = json.loads(reader_metadata.split("\n\n")[1])
        # Move SI dictionary up one level
        if "SI" in metadata_dict.keys():
            si_contents = metadata_dict.pop("SI")
            metadata_dict.update(si_contents)
        return RawImageInfo(
            reader_shape=reader_shape,
            reader_metadata_header=metadata_dict,
            reader_metadata_json=reader_metadata_json,
            reader_descriptions=reader_descriptions,
        )

    @staticmethod
    def map_raw_image_info_to_tif_file_group(
        raw_image_info: RawImageInfo,
    ) -> TifFileGroup:
        """
        Map raw image info to a tiff file group type
        Parameters
        ----------
        raw_image_info : RawImageInfo

        Returns
        -------
        TifFileGroup

        """
        header = raw_image_info.reader_metadata_header
        if header.get("hPhotostim", dict()).get("status") in [
            "'Running'",
            "Running",
        ]:
            return TifFileGroup.PHOTOSTIM
        elif (
            header.get("hIntegrationRoiManager", dict()).get("enable") == "true"
            and header.get("hIntegrationRoiManager", dict()).get("outputChannelsEnabled") == "true"
            and header.get("extTrigEnable", dict()) == "1"
        ):
            return TifFileGroup.BEHAVIOR
        elif header.get("hStackManager", dict()).get("enable") == "true":
            return TifFileGroup.STACK
        else:
            return TifFileGroup.SPONTANEOUS

    def _extract(self, tif_file_locations: Dict[str, List[Path]]) -> ExtractedInfo:
        """
        Loop through list of files and use ScanImageTiffReader to read metadata
        Parameters
        ----------
        tif_file_locations : Dict[str, List[Path]]


        """
        extracted_info = []
        for file_stem, files in tif_file_locations.items():
            last_idx = -1
            metadata_extracted = False
            raw_info = None
            while not metadata_extracted:
                try:
                    last_file = files[last_idx]
                    raw_info = self.extract_raw_info_from_file(last_file)
                    metadata_extracted = True
                except Exception as e:
                    logging.warning(e)
                    last_idx -= 1
            raw_info_first_file = self.extract_raw_info_from_file(files[0])
            tif_file_group = self.map_raw_image_info_to_tif_file_group(raw_image_info=raw_info)
            extracted_item = ExtractedInfoItem(
                raw_info_last_file=raw_info,
                raw_info_first_file=raw_info_first_file,
                file_stem=file_stem,
                tif_file_group=tif_file_group,
                files=files,
            )
            extracted_info.append(extracted_item)

        return ExtractedInfo(info=extracted_info)

    def run_job(self):
        """Main entrypoint to extract info and save to file"""
        tif_file_locations = self.get_tif_file_locations()
        self.metadata = self._extract(tif_file_locations)
        with open(self.settings.output_filepath, "w") as f:
            json.dump(self.metadata.model_dump(mode="json"), f, indent=3)


if __name__ == "__main__":
    sys_args = sys.argv[1:]
    if len(sys_args) == 2 and sys_args[0] == "--settings":
        main_settings = Settings.model_validate_json(sys_args[1])
    else:
        # Pull from env vars and command line args
        # noinspection PyArgumentList
        main_settings = Settings()
    main_job = Extractor(settings=main_settings)
    main_job.run_job()



"""Functions for working with sync files."""

import datetime
import logging
from functools import lru_cache
from typing import Optional, Sequence, Union

import h5py
import numpy as np
import scipy.spatial.distance as distance

import aind_metadata_extractor.utils.camstim_sync.pkl_utils as pkl

logger = logging.getLogger(__name__)


@lru_cache(maxsize=32)
def load_sync(path):
    """
    Loads an hdf5 sync dataset.

    Parameters
    ----------
    path : str
        Path to hdf5 file.

    Returns
    -------
    dfile : h5py.File
        Loaded hdf5 file.
    """
    dfile = h5py.File(path, "r")
    return dfile


def get_meta_data(sync_file):
    """
    Gets the meta data from the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    meta_data : dict
        Meta data from the sync file.
    """
    meta_data = eval(sync_file["meta"][()])
    return meta_data


def get_line_labels(sync_file):
    """
    Gets the line labels from the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    line_labels : list
        Line labels from the sync file.
    """
    meta_data = get_meta_data(sync_file)
    line_labels = meta_data["line_labels"]
    return line_labels


def get_times(sync_file):
    """
    Gets the times from the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    times : np.ndarray
        Times from the sync file.
    """
    times = process_times(sync_file)
    return times


def get_start_time(sync_file) -> datetime.datetime:
    """
    Gets the start timefrom the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    meta_data : dict
        Meta data from the sync file.
    """
    meta_data = get_meta_data(sync_file)
    return datetime.datetime.fromisoformat(meta_data["start_time"])


def get_total_seconds(sync_file) -> float:
    """
    Gets the overall length from the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    data: float
        Total seconds.
    """
    meta_data = get_meta_data(sync_file)
    return meta_data["total_samples"] / get_sample_freq(meta_data)


def get_stop_time(sync_file) -> datetime.datetime:
    """
    Gets the stop time from the sync file.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    data: datetime.datetime
        Stop time.
    """
    start_time = get_start_time(sync_file)
    total_seconds = get_total_seconds(sync_file)
    return start_time + datetime.timedelta(seconds=total_seconds)


def extract_led_times(sync_file, keys="", fallback_line=18):
    """
    Extracts the LED times from the sync file.
    Rising or Falling

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    led_times : np.ndarray
        LED times.
    """

    try:
        led_times = get_edges(sync_file=sync_file, kind="rising", keys=keys, units="seconds")
    except KeyError:
        led_times = get_rising_edges(sync_file, fallback_line, units="seconds")

    return led_times


def process_times(sync_file):
    """
    Processes the times from the sync file.
    Checks for rollover times

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    times : np.ndarray
        Times from the sync file.

    """
    times = sync_file["data"][()][:, 0:1].astype(np.int64)

    intervals = np.ediff1d(times, to_begin=0)
    rollovers = np.where(intervals < 0)[0]

    for i in rollovers:
        times[i:] += 4294967296

    return times


def get_ophys_stimulus_timestamps(sync, pkl):
    """Obtain visual behavior stimuli timing information from a sync *.h5 file.

    Parameters
    ----------
    sync_path : Union[str, Path]
        The path to a sync *.h5 file that contains global timing information
        about multiple data streams (e.g. behavior, ophys, eye_tracking)
        during a session.

    Returns
    -------
    np.ndarray
        Timestamps (in seconds) for presented stimulus frames during a
        behavior + ophys session.
    """
    stimulus_timestamps, _ = get_clipped_stim_timestamps(sync, pkl)
    return stimulus_timestamps


def get_stim_data_length(filename: str) -> int:
    """Get stimulus data length from .pkl file.

    Parameters
    ----------
    filename : str
        Path of stimulus data .pkl file.

    Returns
    -------
    int
        Stimulus data length.
    """
    stim_data = pkl.load_pkl(filename)

    # A subset of stimulus .pkl files do not have the "vsynccount" field.
    # MPE *won't* be backfilling the "vsynccount" field for these .pkl files.
    # So the least worst option is to recalculate the vsync_count.
    try:
        vsync_count = stim_data["vsynccount"]
    except KeyError:
        vsync_count = len(stim_data["items"]["behavior"]["intervalsms"]) + 1

    return vsync_count


def get_behavior_stim_timestamps(sync):
    """
    Get stimulus timestamps from the behavior stream in the sync file.
    Checks various line labels

    Parameters
    ----------
    sync : h5py.File
        Sync file.

    Returns
    -------
    times : np.ndarray
        Timestamps.
    """
    try:
        stim_key = "vsync_stim"
        times = get_falling_edges(sync, stim_key, units="seconds")
        return times
    except ValueError:
        stim_key = "stim_vsync"
        times = get_falling_edges(sync, stim_key, units="seconds")
        return times
    except Exception:
        raise ValueError("No stimulus stream found in sync file")


def get_clipped_stim_timestamps(sync, pkl_path):
    """
    Get stimulus timestamps from the behavior stream in the sync file.
    Checks various line labels
    Clips based on length

    Parameters
    ----------
    sync : h5py.File
        Sync file.
    pkl_path : str
        Path to pkl file

    Returns
    -------
    timestamps : np.ndarray
        Timestamps.
    delta: int
        Difference in length
    """

    timestamps = get_behavior_stim_timestamps(sync)
    stim_data_length = get_stim_data_length(pkl_path)

    delta = 0
    logger.debug(sync)
    if stim_data_length is not None and stim_data_length < len(timestamps):
        try:
            stim_key = "vsync_stim"
            rising = get_rising_edges(sync, stim_key, units="seconds")
        except ValueError:
            stim_key = "stim_vsync"
            rising = get_rising_edges(sync, stim_key, units="seconds")
        except Exception:
            raise ValueError("No stimulus stream found in sync file")

        # Some versions of camstim caused a spike when the DAQ is first
        # initialized. Remove it.
        if rising[1] - rising[0] > 0.2:
            logger.debug("Initial DAQ spike detected from stimulus, " "removing it")
            timestamps = timestamps[1:]

        delta = len(timestamps) - stim_data_length
        if delta != 0:
            logger.debug(
                "Stim data of length %s has timestamps of " "length %s",
                stim_data_length,
                len(timestamps),
            )
    elif stim_data_length is None:
        logger.debug("No data length provided for stim stream")
    return timestamps, delta


def line_to_bit(sync_file, line):
    """
    Returns the bit for a specified line.  Either line name and number is
        accepted.

    Parameters
    ----------
    line : str
        Line name for which to return corresponding bit.

    returns
    -------
    bit : int
        Bit for the line.

    """
    line_labels = get_line_labels(sync_file)

    if type(line) is int:
        return line
    elif type(line) is str:
        return line_labels.index(line)
    else:
        raise TypeError("Incorrect line type.  Try a str or int.")


def get_edges(
    sync_file: h5py.File,
    kind: str,
    keys: Union[str, Sequence[str]],
    units: str = "seconds",
    permissive: bool = False,
) -> Optional[np.ndarray]:
    """
    Utility function for extracting edge times from a line

    Parameters
    ----------
    kind : One of "rising", "falling", or "all". Should this method return
        timestamps for rising, falling or both edges on the appropriate
        line
    keys : These will be checked in sequence. Timestamps will be returned
        for the first which is present in the line labels
    units : one of "seconds", "samples", or "indices". The returned
        "time"stamps will be given in these units.
    raise_missing : If True and no matching line is found, a KeyError will
        be raised

    Returns
    -------
    An array of edge times. If raise_missing is False and none of the keys
        were found, returns None.

    Raises
    ------
    KeyError : none of the provided keys were found among this dataset's
        line labels

    """

    if isinstance(keys, str):
        keys = [keys]

    logger.debug(keys)

    for line in keys:
        try:
            if kind == "falling":
                return get_falling_edges(sync_file, line, units)
            elif kind == "rising":
                return get_rising_edges(sync_file, line, units)
            elif kind == "all":
                return np.sort(
                    np.concatenate(
                        [
                            get_edges(sync_file, "rising", keys, units),
                            get_edges(sync_file, "falling", keys, units),
                        ]
                    )
                )
        except ValueError:
            continue

    if not permissive:
        raise KeyError(f"none of {keys} were found in this dataset's line labels")


def get_bit_changes(sync_file, bit):
    """
    Returns the first derivative of a specific bit.
        Data points are 1 on rising edges and 255 on falling edges.

    Parameters
    ----------
    bit : int
        Bit for which to return changes.

    """
    bit_array = get_sync_file_bit(sync_file, bit)
    return np.ediff1d(bit_array, to_begin=0)


def get_all_bits(sync_file):
    """
    Returns all counter values.

    Parameters
    ----------
    sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    data: np.ndarray
        All counter values.
    """
    return sync_file["data"][()][:, -1]


def get_sync_file_bit(sync_file, bit):
    """
    Returns a specific bit from the sync file.

    Parameters
    ----------
    bit : int
        Bit to extract.
    Sync_file : h5py.File
        Loaded hdf5 file.

    Returns
    -------
    data: np.ndarray
        Bit values.
    """

    return get_bit(get_all_bits(sync_file), bit)


def get_bit(uint_array, bit):
    """
    Returns a bool array for a specific bit in a uint ndarray.

    Parameters
    ----------
    uint_array : (numpy.ndarray)
        The array to extract bits from.
    bit : (int)
        The bit to extract.

    """
    return np.bitwise_and(uint_array, 2**bit).astype(bool).astype(np.uint8)


def get_sample_freq(meta_data):
    """
    Returns the sample frequency from the meta data.

    Parameters
    ----------
    meta_data : dict
        Meta data from the sync file.

    Returns
    -------
    data: float
        Sample frequency.
    """

    try:
        return float(meta_data["ni_daq"]["sample_freq"])
    except KeyError:
        return float(meta_data["ni_daq"]["counter_output_freq"])


def get_all_times(sync_file, meta_data, units="samples"):
    """
    Returns all counter values.

    Parameters
    ----------
    units : str
        Return times in 'samples' or 'seconds'

    """
    if meta_data["ni_daq"]["counter_bits"] == 32:
        times = sync_file["data"][()][:, 0]
    else:
        times = 0
    units = units.lower()
    if units == "samples":
        return times
    elif units in ["seconds", "sec", "secs"]:
        freq = get_sample_freq(meta_data)
        return times / freq
    else:
        raise ValueError("Only 'samples' or 'seconds' are valid units.")


def get_falling_edges(sync_file, line, units="samples"):
    """
    Returns the counter values for the falling edges for a specific bit
        or line.

    Parameters
    ----------
    line : str
        Line for which to return edges.

    """
    meta_data = get_meta_data(sync_file)
    bit = line_to_bit(sync_file, line)
    changes = get_bit_changes(sync_file, bit)
    return get_all_times(sync_file, meta_data, units)[np.where(changes == 255)]


def get_rising_edges(sync_file, line, units="samples"):
    """
    Returns the counter values for the rizing edges for a specific bit or
        line.

    Parameters
    ----------
    line : str
        Line for which to return edges.

    """
    meta_data = get_meta_data(sync_file)
    bit = line_to_bit(sync_file, line)
    changes = get_bit_changes(sync_file, bit)
    return get_all_times(sync_file, meta_data, units)[np.where(changes == 1)]


def trimmed_stats(data, pctiles=(10, 90)):
    """
    Returns the mean and standard deviation of the data after trimming the
        data at the specified percentiles.

    Parameters
    ----------
    data : np.ndarray
        Data to trim.
    pctiles : tuple
        Percentiles at which to trim the data.

    Returns
    -------
    mean : float
        Mean of the trimmed data.
    std : float
        Standard deviation of the trimmed data.
    """
    low = np.percentile(data, pctiles[0])
    high = np.percentile(data, pctiles[1])

    trimmed = data[np.logical_and(data <= high, data >= low)]

    return np.mean(trimmed), np.std(trimmed)


def estimate_frame_duration(pd_times, cycle=60):
    """
    Estimates the frame duration from the photodiode times.

    Parameters
    ----------

    pd_times : np.ndarray
        Photodiode times.
    cycle : int
        Number of frames per cycle.

    Returns
    -------
    frame_duration : float
        Estimated frame duration.
    """
    return trimmed_stats(np.diff(pd_times))[0] / cycle


def allocate_by_vsync(vs_diff, index, starts, ends, frame_duration, irregularity, cycle):
    """
    Allocates frame times based on the vsync signal.

    Parameters
    ----------
    vs_diff : np.ndarray
        Difference between vsync times.
    index : int
        Index of the current vsync.
    starts : np.ndarray
        Start times of the frames.
    ends : np.ndarray
        End times of the frames.
    frame_duration : float
        Duration of the frame.
    irregularity : int
        Irregularity in the frame times.
    cycle : int
        Number of frames per cycle.

    Returns
    -------
    starts : np.ndarray
        Start times of the frames.
    ends : np.ndarray
        End times of the frames.
    """
    current_vs_diff = vs_diff[index * cycle : (index + 1) * cycle]
    sign = np.sign(irregularity)

    if sign > 0:
        vs_ind = np.argmax(current_vs_diff)
    elif sign < 0:
        vs_ind = np.argmin(current_vs_diff)

    ends[vs_ind:] += sign * frame_duration
    starts[vs_ind + 1 :] += sign * frame_duration

    return starts, ends


def trim_border_pulses(pd_times, vs_times, frame_interval=1 / 60, num_frames=5):
    """
    Trims pulses near borders of the photodiode signal.

    Parameters
    ----------
    pd_times : np.ndarray
        Photodiode times.
    vs_times : np.ndarray
        Vsync times.
    frame_interval : float
        Interval between frames.
    num_frames : int
        Number of frames.

    Returns
    -------
    pd_times : np.ndarray
        Trimmed photodiode times.
    """
    pd_times = np.array(pd_times)
    return pd_times[
        np.logical_and(
            pd_times >= vs_times[0],
            pd_times <= vs_times[-1] + num_frames * frame_interval,
        )
    ]


def correct_on_off_effects(pd_times):
    """

    Notes
    -----
    This cannot (without additional info) determine whether an assymmetric
    offset is odd-long or even-long.
    """

    pd_diff = np.diff(pd_times)
    odd_diff_mean, odd_diff_std = trimmed_stats(pd_diff[1::2])
    even_diff_mean, even_diff_std = trimmed_stats(pd_diff[0::2])

    half_diff = np.diff(pd_times[0::2])
    full_period_mean, full_period_std = trimmed_stats(half_diff)
    half_period_mean = full_period_mean / 2

    odd_offset = odd_diff_mean - half_period_mean
    even_offset = even_diff_mean - half_period_mean

    pd_times[::2] -= odd_offset / 2
    pd_times[1::2] -= even_offset / 2

    return pd_times


def trim_discontiguous_vsyncs(vs_times, photodiode_cycle=60):
    """
    Trims discontiguous vsyncs from the photodiode signal.

    Parameters
    ----------
    vs_times : np.ndarray
        Vsync times.
    photodiode_cycle : int
        Number of frames per cycle.

    Returns
    -------
    vs_times : np.ndarray
        Trimmed vsync times.
    """
    vs_times = np.array(vs_times)

    breaks = np.where(np.diff(vs_times) > (1 / photodiode_cycle) * 100)[0]

    if len(breaks) > 0:
        chunk_sizes = np.diff(
            np.concatenate(
                (
                    np.array(
                        [
                            0,
                        ]
                    ),
                    breaks,
                    np.array(
                        [
                            len(vs_times),
                        ]
                    ),
                )
            )
        )
        largest_chunk = np.argmax(chunk_sizes)

        if largest_chunk == 0:
            return vs_times[: np.min(breaks + 1)]
        elif largest_chunk == len(breaks):
            return vs_times[np.max(breaks + 1) :]
        else:
            return vs_times[breaks[largest_chunk - 1] : breaks[largest_chunk]]
    else:
        return vs_times


def assign_to_last(starts, ends, frame_duration, irregularity):
    """
    Assigns the irregularity to the last frame.

    Parameters
    ----------
    starts : np.ndarray
        Start times of the frames.
    ends : np.ndarray
        End times of the frames.
    frame_duration : float
        Duration of the frame.
    irregularity : int
        Irregularity in the frame times.

    Returns
    -------
    starts : np.ndarray
        Start times of the frames.
    ends : np.ndarray
        Modified end times of the frames.
    """
    ends[-1] += frame_duration * np.sign(irregularity)
    return starts, ends


def remove_zero_frames(frame_times):
    """
    Removes zero delta frames from the frame times.

    Parameters
    ----------
    frame_times : np.ndarray
        Frame times.

    Returns
    -------
    t : np.ndarray
        Modified frame times.
    """
    deltas = np.diff(frame_times)

    small_deltas = np.where(deltas < 0.01)[0]
    big_deltas = np.where((deltas > 0.018) * (deltas < 0.1))[0]

    def find_match(big_deltas, value):
        """
        Finds max match for the value in the big deltas.

        Parameters
        ----------
        big_deltas : np.ndarray
            Big deltas.
        value : float
            Value to match.

        Returns
        -------
        float
            Matched value.
        """

        try:
            return big_deltas[np.max(np.where((big_deltas < value))[0])] - value
        except ValueError:
            return None

    paired_deltas = [find_match(big_deltas, A) for A in small_deltas]

    ft = np.copy(deltas)

    for idx, d in enumerate(small_deltas):
        if paired_deltas[idx] is not None:
            if paired_deltas[idx] > -100:
                ft[d + paired_deltas[idx]] = np.median(deltas)
                ft[d] = np.median(deltas)

    t = np.concatenate(([np.min(frame_times)], np.cumsum(ft) + np.min(frame_times)))

    return t


def compute_frame_times(
    photodiode_times,
    frame_duration,
    num_frames,
    cycle,
    irregular_interval_policy=assign_to_last,
):
    """
    Computes the frame times from the photodiode times.

    Parameters
    ----------
    photodiode_times : np.ndarray
        Photodiode times.
    frame_duration : float
        Duration of the frame.
    num_frames : int
        Number of frames.
    cycle : int
        Number of frames per cycle.
    irregular_interval_policy : function
        Policy for handling irregular intervals.

    Returns
    -------
    indices : np.ndarray
        Indices of the frames.
    starts : np.ndarray
        Start times of the frames.
    ends : np.ndarray
        End times of the frames.
    """
    indices = np.arange(num_frames)
    starts = np.zeros(num_frames, dtype=float)
    ends = np.zeros(num_frames, dtype=float)

    num_intervals = len(photodiode_times) - 1
    for start_index, (start_time, end_time) in enumerate(zip(photodiode_times[:-1], photodiode_times[1:])):
        interval_duration = end_time - start_time
        irregularity = int(np.around((interval_duration) / frame_duration)) - cycle

        local_frame_duration = interval_duration / (cycle + irregularity)
        durations = np.zeros(cycle + (start_index == num_intervals - 1)) + local_frame_duration

        current_ends = np.cumsum(durations) + start_time
        current_starts = current_ends - durations

        while irregularity != 0:
            current_starts, current_ends = irregular_interval_policy(
                start_index,
                current_starts,
                current_ends,
                local_frame_duration,
                irregularity,
                cycle,
            )
            irregularity += -1 * np.sign(irregularity)

        early_frame = start_index * cycle
        late_frame = (start_index + 1) * cycle + (start_index == num_intervals - 1)

        remaining = starts[early_frame:late_frame].size
        starts[early_frame:late_frame] = current_starts[:remaining]
        ends[early_frame:late_frame] = current_ends[:remaining]

    return indices, starts, ends


def separate_vsyncs_and_photodiode_times(vs_times, pd_times, photodiode_cycle=60):
    """
    Separates the vsyncs and photodiode times.

    Parameters
    ----------
    vs_times : np.ndarray
        Vsync times.
    pd_times : np.ndarray
        Photodiode times.

    Returns
    -------
    vs_times_out : np.ndarray
        Vsync times.
    pd_times_out : np.ndarray
        Photodiode times.
    """
    vs_times = np.array(vs_times)
    pd_times = np.array(pd_times)

    breaks = np.where(np.diff(vs_times) > (1 / photodiode_cycle) * 100)[0]

    shift = 2.0
    break_times = [-shift]
    break_times.extend(vs_times[breaks].tolist())
    break_times.extend([np.inf])

    vs_times_out = []
    pd_times_out = []

    for indx, b in enumerate(break_times[:-1]):
        pd_in_range = np.where((pd_times > break_times[indx] + shift) * (pd_times <= break_times[indx + 1] + shift))[0]
        vs_in_range = np.where((vs_times > break_times[indx]) * (vs_times <= break_times[indx + 1]))[0]

        vs_times_out.append(vs_times[vs_in_range])
        pd_times_out.append(pd_times[pd_in_range])

    return vs_times_out, pd_times_out


def flag_unexpected_edges(pd_times, ndevs=10):
    """
    Flags unexpected edges in the photodiode times.

    Parameters
    ----------
    pd_times : np.ndarray
        Photodiode times.
    ndevs : int
        Number of standard deviations.

    Returns
    -------
    expected_duration_mask : np.ndarray
        Mask for expected durations.
    """
    pd_diff = np.diff(pd_times)
    diff_mean, diff_std = trimmed_stats(pd_diff)

    expected_duration_mask = np.ones(pd_diff.size)
    expected_duration_mask[
        np.logical_or(
            pd_diff < diff_mean - ndevs * diff_std,
            pd_diff > diff_mean + ndevs * diff_std,
        )
    ] = 0
    expected_duration_mask[1:] = np.logical_and(expected_duration_mask[:-1], expected_duration_mask[1:])
    expected_duration_mask = np.concatenate([expected_duration_mask, [expected_duration_mask[-1]]])

    return expected_duration_mask


def fix_unexpected_edges(pd_times, ndevs=10, cycle=60, max_frame_offset=4):
    """
    Fixes unexpected edges in the photodiode times.

    Parameters
    ----------
    pd_times : np.ndarray
        Photodiode times.
    ndevs : int
        Number of standard deviations.
    cycle : int
        Number of frames per cycle.
    max_frame_offset : int
        Maximum frame offset.

    Returns
    -------
    output_edges : np.ndarray
        Output edges.
    """
    pd_times = np.array(pd_times)
    expected_duration_mask = flag_unexpected_edges(pd_times, ndevs=ndevs)
    diff_mean, diff_std = trimmed_stats(np.diff(pd_times))
    frame_interval = diff_mean / cycle

    bad_edges = np.where(expected_duration_mask == 0)[0]
    bad_blocks = np.sort(
        np.unique(
            np.concatenate(
                [
                    [0],
                    np.where(np.diff(bad_edges) > 1)[0] + 1,
                    [len(bad_edges)],
                ]
            )
        )
    )

    output_edges = []
    for low, high in zip(bad_blocks[:-1], bad_blocks[1:]):
        current_bad_edge_indices = bad_edges[low : high - 1]
        current_bad_edges = pd_times[current_bad_edge_indices]
        low_bound = pd_times[current_bad_edge_indices[0]]
        high_bound = pd_times[current_bad_edge_indices[-1] + 1]

        edges_missing = int(np.around((high_bound - low_bound) / diff_mean))
        expected = np.linspace(low_bound, high_bound, edges_missing + 1)

        distances = distance.cdist(current_bad_edges[:, None], expected[:, None])
        distances = np.around(distances / frame_interval).astype(int)

        min_offsets = np.amin(distances, axis=0)
        min_offset_indices = np.argmin(distances, axis=0)
        output_edges = np.concatenate(
            [
                output_edges,
                expected[min_offsets > max_frame_offset],
                current_bad_edges[min_offset_indices[min_offsets <= max_frame_offset]],
            ]
        )

    return np.sort(np.concatenate([output_edges, pd_times[expected_duration_mask > 0]]))



"""Mesoscope metadata model"""

from pydantic import BaseModel, Field


class MesoscopeExtractModel(BaseModel):
    """Mesoscope model for extracting metadata."""

    tiff_header: list = Field(title="Header information from TIFF files")
    session_metadata: dict = Field(title="Metadata extracted from the session platform JSON and other data")
    camstim_epchs: list = Field(title="List of epochs from the Camstim platform")
    camstim_session_type: str = Field(title="Type of session from the Camstim platform")
    job_settings: dict = Field(title="Job settings used for extraction")



"""Utility functions for SmartSPIM metadata extraction."""

from datetime import datetime
import os
import json
import re
from typing import Any, List, Optional


def read_json_as_dict(filepath: str) -> dict:
    """
    Reads a json as dictionary with robust encoding handling.
    Parameters
    ------------------------
    filepath: PathLike
        Path where the json is located.
    Returns
    ------------------------
    dict:
        Dictionary with the data the json has.
    """

    dictionary = {}

    if os.path.exists(filepath):
        try:
            with open(filepath, "r", encoding="utf-8") as json_file:
                dictionary = json.load(json_file)
        except (UnicodeDecodeError, json.JSONDecodeError) as e:
            print(f"Error reading json with utf-8: {e}")
            print("Falling back to binary read with character replacement.")

            # Fallback: read as binary and replace problematic characters
            with open(filepath, "rb") as json_file:
                data = json_file.read()
                # Decode with replacement, then clean up common problematic characters
                data_str = data.decode("utf-8", errors="replace")
                # Replace common non-UTF-8 micrometer symbols with proper UTF-8 µ
                data_str = data_str.replace("�m", "um")  # Replace replacement char + m with µm
                data_str = data_str.replace("μm", "um")  # Replace Greek mu with micro sign
                dictionary = json.loads(data_str)

    return dictionary


def get_anatomical_direction(anatomical_direction: str) -> str:
    """
    This function returns the correct anatomical
    direction defined in the aind_data_schema.

    Parameters
    ----------
    anatomical_direction: str
        String defining the anatomical direction
        of the data

    Returns
    -------
    AnatomicalDirection: class::Enum
        Corresponding enum defined in the anatomical
        direction class
    """
    anatomical_direction = anatomical_direction.strip().lower().replace(" ", "_")

    return anatomical_direction


def digest_asi_line(line: str) -> Optional[datetime]:
    """
    Scrape a datetime from a non-empty line, otherwise return None

    Parameters
    -----------
    line: str
        Line from the ASI file

    Returns
    -----------
    datetime
        A date that could be parsed from a string
    """

    if line.isspace():
        return None

    try:
        parts = line.split()
        if len(parts) < 3:
            return None

        mdy, hms, ampm = parts[0:3]

        mdy_parts = [int(i) for i in mdy.split("/")]
        ymd = [mdy_parts[i] for i in [2, 0, 1]]

        hms_parts = [int(i) for i in hms.split(":")]

        # Handle AM/PM conversion
        if ampm == "PM" and hms_parts[0] != 12:
            hms_parts[0] += 12
        elif ampm == "AM" and hms_parts[0] == 12:
            hms_parts[0] = 0

        ymdhms = ymd + hms_parts

        dtime = datetime(*ymdhms)
        return dtime
    except (ValueError, IndexError):
        # Return None for lines that can't be parsed as timestamps
        return None


def get_session_end(asi_file: os.PathLike) -> datetime:
    """
    Work backward from the last line until there is a timestamp

    Parameters
    ------------
    asi_file: PathLike
        Path where the ASI metadata file is
        located

    Returns
    ------------
    Date when the session ended
    """

    with open(asi_file, "rb") as file:
        asi_mdata = file.readlines()

    idx = -1
    result = None
    while result is None:
        result = digest_asi_line(asi_mdata[idx].decode())
        idx -= 1

    return result


def get_excitation_emission_waves(channels: List) -> dict:
    """
    Gets the excitation and emission waves for
    the existing channels within a dataset

    Parameters
    ------------
    channels: List[str]
        List with the channels.
        They must contain the emmision
        wavelenght in the name

    Returns
    ------------
    dict
        Dictionary with the excitation
        and emission waves
    """
    excitation_emission_channels = {}

    for channel in channels:
        channel = channel.replace("Em_", "").replace("Ex_", "")
        splitted = channel.split("_")
        excitation_emission_channels[splitted[0]] = int(splitted[1])

    return excitation_emission_channels


def parse_channel_name(channel_str: str) -> str:
    """
    Parses the channel string from SLIMS to a standard format.

    Parameters
    ----------
    channel_str: str
        The channel name to be parsed.
          ex: "Laser = 445; Emission Filter = 469/35"
          ex: "Laser = 488, Emission Filter = 525/50"
    Returns
    -------
    str
        The parsed channel name (ex: "Ex_445_Em_469").
    """
    s = channel_str.replace("Laser", "Ex").replace("Emission Filter", "Em")
    parts = [p.strip() for p in re.split(r"[;,]", s) if p.strip()]
    segments = []
    for part in parts:
        key, val = [t.strip() for t in part.split("=", 1)]
        # discard any bandwidth info after slash
        core = val.split("/", 1)[0]
        segments.append(f"{key}_{core}")

    return "_".join(segments)


def ensure_list(raw: Any) -> List[Any]:
    """
    Turn a value that might be a list, a single string, or None
    into a proper list of strings (or an empty list).
    """
    if isinstance(raw, list):
        return raw
    if isinstance(raw, str) and raw.strip():
        return [raw]
    return []



"""Module defining Settings for Bergamo Extractor class."""

from decimal import Decimal
from pathlib import Path
from typing import List, Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings, cli_parse_args=True, cli_ignore_unknown_args=True):
    """
    Data that needs to be input by user. Can be pulled from env vars with BERGAMO prefix or set explicitly.
    """

    model_config = SettingsConfigDict(env_prefix="BERGAMO_")
    input_source: Path = Field(..., description="Path to input source.")
    output_filepath: Path = Field(..., description=("Dump model to json file if set."))
    # mandatory fields:
    experimenter_full_name: List[str]
    subject_id: str
    imaging_laser_wavelength: int  # user defined
    fov_imaging_depth: int
    fov_targeted_structure: str
    notes: Optional[str]

    # fields with default values
    mouse_platform_name: str = "Standard Mouse Tube"  # FROM RIG JSON
    active_mouse_platform: bool = False
    session_type: str = "BCI"
    iacuc_protocol: str = "2109"
    # should match rig json:
    rig_id: str = "442 Bergamo 2p photostim"
    behavior_camera_names: List[str] = [
        "Side Face Camera",
        "Bottom Face Camera",
    ]
    ch1_filter_names: List[str] = [
        "Green emission filter",
        "Emission dichroic",
    ]
    ch1_detector_name: str = "Green PMT"
    ch1_daq_name: str = "PXI"
    ch2_filter_names: List[str] = ["Red emission filter", "Emission dichroic"]
    ch2_detector_name: str = "Red PMT"
    ch2_daq_name: str = "PXI"
    imaging_laser_name: str = "Chameleon Laser"

    photostim_laser_name: str = "Monaco Laser"
    stimulus_device_names: List[str] = ["speaker", "lickport"]  # FROM RIG JSON
    photostim_laser_wavelength: int = 1040
    fov_coordinate_ml: Decimal = Decimal("1.5")
    fov_coordinate_ap: float = Decimal("1.5")
    fov_reference: str = "Bregma"

    starting_lickport_position: List[float] = [
        0,
        -6,
        0,
    ]  # in mm from face of the mouse
    behavior_task_name: str = "single neuron BCI conditioning"
    hit_rate_trials_0_10: Optional[float] = None
    hit_rate_trials_20_40: Optional[float] = None
    total_hits: Optional[float] = None
    average_hit_rate: Optional[float] = None
    trial_num: Optional[float] = None
    # ZoneInfo object doesn't serialize well, so we can define it as a str
    timezone: str = "US/Pacific"



"""SmartSPIM extractor module"""

from datetime import datetime
import os
import re
from typing import Optional
from pathlib import Path

import requests

from aind_metadata_extractor.core import BaseExtractor
from aind_metadata_extractor.smartspim.job_settings import JobSettings
from aind_metadata_extractor.models.smartspim import SmartspimModel, FileMetadataModel, SlimsMetadataModel
from aind_metadata_extractor.smartspim.utils import get_excitation_emission_waves, get_session_end, read_json_as_dict

REGEX_DATE = r"(20[0-9]{2})-([0-9]{2})-([0-9]{2})_([0-9]{2})-" r"([0-9]{2})-([0-9]{2})"
REGEX_MOUSE_ID = r"([0-9]{6})"


class SmartspimExtractor(BaseExtractor):
    """Extractor for SmartSPIM metadata from microscope files and SLIMS."""

    def __init__(self, job_settings: JobSettings):
        """Initialize the SmartSPIM extractor with job settings."""
        self.metadata = None
        self.job_settings = job_settings

    def run_job(self) -> dict:
        """Run the extraction job."""
        self.metadata = self._extract()
        return self.metadata.model_dump()

    def _extract(self) -> SmartspimModel:
        """Run extraction process"""

        file_metadata = self._extract_metadata_from_microscope_files()
        slims_metadata = self._extract_metadata_from_slims()

        # Create model objects
        file_metadata_model = FileMetadataModel(**file_metadata)
        slims_metadata_model = SlimsMetadataModel(**slims_metadata)

        smartspim_metadata = SmartspimModel(
            acquisition_type=self.job_settings.acquisition_type,
            file_metadata=file_metadata_model,
            slims_metadata=slims_metadata_model,
        )

        return smartspim_metadata

    def _extract_metadata_from_microscope_files(self) -> dict:
        """
        Extracts metadata from the microscope metadata files.

        Returns
        -------
        Dict
            Dictionary containing metadata from
            the microscope for the current acquisition. This
            is needed to build the acquisition.json.
        """
        # Convert input_source to Path - handle various input types
        if isinstance(self.job_settings.input_source, (str, Path)):
            input_path = Path(self.job_settings.input_source)
        elif isinstance(self.job_settings.input_source, list) and len(self.job_settings.input_source) > 0:
            # Take the first path if it's a list
            input_path = Path(self.job_settings.input_source[0])
        else:
            raise ValueError("input_source must be a valid path or list of paths")

        # Path where the channels are stored
        smartspim_channel_root = input_path.joinpath("SmartSPIM")

        # Getting only valid folders
        channels = [
            folder
            for folder in os.listdir(smartspim_channel_root)
            if os.path.isdir(f"{smartspim_channel_root}/{folder}")
        ]

        # Path to metadata files
        asi_file_path_txt = input_path.joinpath(self.job_settings.asi_filename)

        mdata_path = input_path.joinpath(self.job_settings.mdata_filename_json)

        # ASI file does not exist, needed for acquisition
        if not asi_file_path_txt.exists():
            raise FileNotFoundError(f"File {asi_file_path_txt} does not exist")

        if not mdata_path.exists():
            raise FileNotFoundError(f"File {mdata_path} does not exist")

        # Getting acquisition metadata from the microscope
        metadata_info = read_json_as_dict(str(mdata_path))

        filter_mapping = get_excitation_emission_waves(channels)
        session_config = metadata_info["session_config"]
        wavelength_config = metadata_info["wavelength_config"]
        tile_config = metadata_info["tile_config"]

        if None in [session_config, wavelength_config, tile_config]:
            raise ValueError("Metadata json is empty")

        session_end_time = get_session_end(asi_file_path_txt)
        mdate_match = re.search(REGEX_DATE, input_path.stem)
        if not (mdate_match):
            raise ValueError("Error while extracting session date.")
        session_start = datetime.strptime(mdate_match.group(), "%Y-%m-%d_%H-%M-%S")

        metadata_dict = {
            "session_config": session_config,
            "wavelength_config": wavelength_config,
            "tile_config": tile_config,
            "session_start_time": session_start,
            "session_end_time": session_end_time,
            "filter_mapping": filter_mapping,
        }

        return metadata_dict

    def _extract_metadata_from_slims(
        self, start_date_gte: Optional[str] = None, end_date_lte: Optional[str] = None
    ) -> dict:
        """
        Method to retrieve smartspim imaging info from SLIMS
        using the metadata service endpoint.
        Parameters
        ----------
        start_date_gte: str
            Start date for the search.
        end_date_lte: str
            End date for the search.
        Returns
        -------
        Dict
            Dictionary containing metadata from SLIMS for an acquisition.
        """
        query_params = {"subject_id": self.job_settings.subject_id}
        query_params["start_date_gte"] = start_date_gte if start_date_gte else "2020-01-01"
        query_params["end_date_lte"] = end_date_lte if end_date_lte else "2100-01-01"

        response = requests.get(
            f"{self.job_settings.metadata_service_path}",
            params=query_params,
        )
        response.raise_for_status()
        response_data = response.json().get("data", [])
        if response.status_code == 200 and len(response_data) > 1:
            raise ValueError(
                "More than one imaging session found for the same subject_id. " "Please refine your search."
            )
        elif response.status_code == 200 and len(response_data) == 1:
            imaging_info = response_data[0]
        else:
            raise ValueError("No imaging session found for the given subject_id and date range.")
        return imaging_info



"""Utils to process behavior info for stimulus"""

import logging
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd

import aind_metadata_extractor.utils.camstim_sync.constants as constants
import aind_metadata_extractor.utils.camstim_sync.naming_utils as names
import aind_metadata_extractor.utils.camstim_sync.pkl_utils as pkl
import aind_metadata_extractor.utils.camstim_sync.stim_utils as stim

INT_NULL = -99

logger = logging.getLogger(__name__)


def get_stimulus_presentations(data, stimulus_timestamps) -> pd.DataFrame:
    """
    This function retrieves the stimulus presentation dataframe and
    renames the columns, adds a stop_time column, and set's index to
    stimulus_presentation_id before sorting and returning the dataframe.
    :param data: stimulus file associated with experiment id
    :param stimulus_timestamps: timestamps indicating when stimuli switched
                                during experiment
    :return: stimulus_table: dataframe containing the stimuli metadata as well
                             as what stimuli was presented
    """
    stimulus_table = get_visual_stimuli_df(data, stimulus_timestamps)
    # workaround to rename columns to harmonize with visual
    # coding and rebase timestamps to sync time
    stimulus_table.insert(loc=0, column="flash_number", value=np.arange(0, len(stimulus_table)))
    stimulus_table = stimulus_table.rename(
        columns={
            "frame": "start_frame",
            "time": "start_time",
            "flash_number": "stimulus_presentations_id",
        }
    )
    stimulus_table.start_time = [
        stimulus_timestamps[int(start_frame)] for start_frame in stimulus_table.start_frame.values
    ]
    end_time = []
    logger.debug(f"stimulus_table {stimulus_table}")
    for end_frame in stimulus_table.end_frame.values:
        if not np.isnan(end_frame):
            end_time.append(stimulus_timestamps[int(end_frame)])
        else:
            end_time.append(float("nan"))

    stimulus_table.insert(loc=4, column="stop_time", value=end_time)
    stimulus_table.set_index("stimulus_presentations_id", inplace=True)
    stimulus_table = stimulus_table[sorted(stimulus_table.columns)]
    return stimulus_table


def remove_short_sandwiched_spontaneous(df, duration_thresh=(0.3, 0.5)):
    """
    Removes spontaneous intervals of short duration that are
        sandwiched between two identical stimulus intervals.
    Used to prevent the gray screens during these period
        for loading images being "spontaneous"

    Parameters
    ----------
    df : pandas.DataFrame
        A DataFrame containing stimulus presentation intervals

    duration_thresh : tuple of float, optional
        A (min, max) tuple specifying the duration range
            (in seconds) for spontaneous intervals that
        are candidates for removal. Default is (0.3, 0.5).

    Returns
    -------
    pandas.DataFrame
        A cleaned DataFrame with short,
            sandwiched spontaneous intervals removed.
    """
    # Sort by start_time to ensure order
    df = df.sort_values(by="start_time").reset_index(drop=True)

    # Find spontaneous rows
    is_spont = df["image_name"] == "spontaneous"
    duration_ok = df["duration"].between(*duration_thresh)

    # Iterate over spontaneous rows with short duration
    drop_indices = []
    for idx in df[is_spont & duration_ok].index:
        if idx == 0 or idx == len(df) - 1:
            continue  # Can't check neighbors if on edge

        prev_row = df.loc[idx - 1]
        next_row = df.loc[idx + 1]

        if (
            prev_row["image_name"] == next_row["image_name"]
            and prev_row["stim_name"] == next_row["stim_name"]
            and prev_row["image_name"] != "spontaneous"
        ):
            drop_indices.append(idx)

    return df.drop(index=drop_indices).reset_index(drop=True)


def get_images_dict(pkl_dict) -> Dict:
    """
    Gets the dictionary of images that were presented during an experiment
    along with image set metadata and the image specific metadata. This
    function uses the path to the image pkl file to read the images and their
    metadata from the pkl file and return this dictionary.
    Parameters
    ----------
    pkl_dict: The pkl file containing the data for the stimuli presented during
         experiment

    Returns
    -------
    Dict:
        A dictionary containing keys images, metadata, and image_attributes.
        These correspond to paths to image arrays presented, metadata
        on the whole set of images, and metadata on specific images,
        respectively.

    """
    # Sometimes the source is a zipped pickle:
    pkl_stimuli = pkl_dict["items"]["behavior"]["stimuli"]
    metadata = {"image_set": pkl_stimuli["images"]["image_path"]}

    # Get image file name;
    # These are encoded case-insensitive in the pickle file :/
    filename = stim.convert_filepath_caseinsensitive(metadata["image_set"])

    image_set = pkl.load_img_pkl(open(filename, "rb"))
    images = []
    images_meta = []

    for cat, cat_images in image_set.items():
        for img_index, (img_name, img) in enumerate(cat_images.items()):
            meta = {
                "image_category": cat,
                "image_name": img_name,
                "orientation": np.nan,
                "phase": np.nan,
                "size": np.nan,
                "spatial_frequency": np.nan,
                "image_index": img_index,
            }

            images.append(img)
            images_meta.append(meta)

    images_dict = dict(
        metadata=metadata,
        images=images,
        image_attributes=images_meta,
    )

    return images_dict


def get_stimulus_metadata(pkl) -> pd.DataFrame:
    """
    Gets the stimulus metadata for each type of stimulus presented during
    the experiment. The metadata is return for gratings, images, and omitted
    stimuli.
    Parameters
    ----------
    pkl: the pkl file containing the information about what stimuli were
         presented during the experiment

    Returns
    -------
    pd.DataFrame:
        The dataframe containing a row for every stimulus that was presented
        during the experiment. The row contains the following data,
        image_category, image_name, image_set, phase, spatial_frequency,
        orientation, and image index.

    """
    stimuli = pkl["items"]["behavior"]["stimuli"]
    if "images" in stimuli:
        images = get_images_dict(pkl)
        stimulus_index_df = pd.DataFrame(images["image_attributes"])
        image_set_filename = stim.convert_filepath_caseinsensitive(images["metadata"]["image_set"])
        stimulus_index_df["image_set"] = stim.get_image_set_name(image_set_path=image_set_filename)
    else:
        stimulus_index_df = pd.DataFrame(
            columns=[
                "image_name",
                "image_category",
                "image_set",
                "phase",
                "size",
                "orientation",
                "spatial_frequency",
                "image_index",
            ]
        )
        stimulus_index_df = stimulus_index_df.astype(
            {
                "image_name": str,
                "image_category": str,
                "image_set": str,
                "phase": float,
                "size": float,
                "orientation": float,
                "spatial_frequency": float,
                "image_index": int,
            }
        )

    # get the grating metadata will be empty if gratings are absent
    # grating_df = get_gratings_metadata(
    #    stimuli, start_idx=len(stimulus_index_df)
    # )
    # stimulus_index_df = pd.concat(
    #    [stimulus_index_df, grating_df], ignore_index=True, sort=False
    # )
    # Add an entry for omitted stimuli
    omitted_df = pd.DataFrame(
        {
            "image_category": ["omitted"],
            "image_name": ["omitted"],
            "image_set": ["omitted"],
            "orientation": np.nan,
            "phase": np.nan,
            "size": np.nan,
            "spatial_frequency": np.nan,
            "image_index": len(stimulus_index_df),
        }
    )
    stimulus_index_df = pd.concat([stimulus_index_df, omitted_df], ignore_index=True, sort=False)
    stimulus_index_df.set_index(["image_index"], inplace=True, drop=True)
    # print(stimulus_index_df.head(100))
    return stimulus_index_df


def get_stimulus_epoch(
    set_log: List[Tuple[str, Union[str, int], int, int]],
    current_set_index: int,
    start_frame: int,
    n_frames: int,
) -> Tuple[int, int]:
    """
    Gets the frame range for which a stimuli was presented and the transition
    to the next stimuli was ongoing. Returns this in the form of a tuple.
    Parameters
    ----------
    set_log: List[Tuple[str, Union[str, int], int, int
        The List of Tuples in the form of
        (stimuli_type ('Image' or 'Grating'),
         stimuli_descriptor (image_name or orientation of grating in degrees),
         nonsynced_time_of_display (not sure, it's never used),
         display_frame (frame that stimuli was displayed))
    current_set_index: int
        Index of stimuli set to calculate window
    start_frame: int
        frame where stimuli was set, set_log[current_set_index][3]
    n_frames: int
        number of frames for which stimuli were displayed

    Returns
    -------
    Tuple[int, int]:
        A tuple where index 0 is start frame of stimulus window and index 1 is
        end frame of stimulus window

    """
    try:
        next_set_event = set_log[current_set_index + 1]
    except IndexError:  # assume this is the last set event
        next_set_event = (
            None,
            None,
            None,
            n_frames,
        )

    return start_frame, next_set_event[3]  # end frame isn't inclusive


def get_draw_epochs(draw_log: List[int], start_frame: int, stop_frame: int) -> List[Tuple[int, int]]:
    """
    Gets the frame numbers of the active frames within a stimulus window.
    Stimulus epochs come in the form [0, 0, 1, 1, 0, 0] where the stimulus is
    active for some amount of time in the window indicated by int 1 at that
    frame. This function returns the ranges for which the set_log is 1 within
    the draw_log window.
    Parameters
    ----------
    draw_log: List[int]
        A list of ints indicating for what frames stimuli were active
    start_frame: int
        The start frame to search within the draw_log for active values
    stop_frame: int
        The end frame to search within the draw_log for active values

    Returns
    -------
    List[Tuple[int, int]]
        A list of tuples indicating the start and end frames of every
        contiguous set of active values within the specified window
        of the draw log.
    """
    draw_epochs = []
    current_frame = start_frame

    while current_frame <= stop_frame:
        epoch_length = 0
        while current_frame < stop_frame and draw_log[current_frame] == 1:
            epoch_length += 1
            current_frame += 1
        else:
            current_frame += 1

        if epoch_length:
            draw_epochs.append(
                (
                    current_frame - epoch_length - 1,
                    current_frame - 1,
                )
            )

    return draw_epochs


def unpack_change_log(change):
    """
    Unpacks the change log into a dictionary containing the category of the
    stimuli that was changed, the name of the stimuli that was changed, the
    category of the stimuli that the change was made to, the name of the
    stimuli that the change was made to, the time of the change, and the frame
    of the change.

    Parameters
    ----------
    change: Tuple[str, str, str, str, int, int]
        A tuple containing the category of the stimuli that was changed

    Returns
    -------
    Dict:
        A dictionary containing the category of the stimuli that was changed,
        the name of the stimuli that was changed, the category of the stimuli
        that the change was made to, the name of the stimuli that the change
        was made to, the time of the change, and the frame of the change.
    """

    (
        (from_category, from_name),
        (
            to_category,
            to_name,
        ),
        time,
        frame,
    ) = change

    return dict(
        frame=frame,
        time=time,
        from_category=from_category,
        to_category=to_category,
        from_name=from_name,
        to_name=to_name,
    )


def get_visual_stimuli_df(data, time) -> pd.DataFrame:
    """
    This function loads the stimuli and the omitted stimuli into a dataframe.
    These stimuli are loaded from the input data, where the set_log and
    draw_log contained within are used to calculate the epochs. These epochs
    are used as start_frame and end_frame and converted to times by input
    stimulus timestamps. The omitted stimuli do not have a end_frame by design
    though there duration is always 250ms.
    :param data: the behavior data file
    :param time: the stimulus timestamps indicating when each stimuli is
                 displayed
    :return: df: a pandas dataframe containing the stimuli and omitted stimuli
                 that were displayed with their frame, end_frame, start_time,
                 and duration
    """
    try:
        stimuli = data["items"]["behavior"]["stimuli"]
    except KeyError:
        stimuli = data["items"]["foraging"]["stimuli"]
    n_frames = len(time)
    visual_stimuli_data = []
    for stim_dict in stimuli.values():
        for idx, (attr_name, attr_value, _, frame) in enumerate(stim_dict["set_log"]):
            orientation = attr_value if attr_name.lower() == "ori" else np.nan
            image_name = attr_value if attr_name.lower() == "image" else np.nan

            stimulus_epoch = get_stimulus_epoch(
                stim_dict["set_log"],
                idx,
                frame,
                n_frames,
            )
            draw_epochs = get_draw_epochs(stim_dict["draw_log"], *stimulus_epoch)

            for epoch_start, epoch_end in draw_epochs:
                visual_stimuli_data.append(
                    {
                        "orientation": orientation,
                        "image_name": image_name,
                        "frame": epoch_start,
                        "end_frame": epoch_end,
                        "time": time[epoch_start],
                        "duration": time[epoch_end] - time[epoch_start],
                        # this will always work because an epoch
                        # will never occur near the end of time
                        "omitted": False,
                    }
                )

    visual_stimuli_df = pd.DataFrame(data=visual_stimuli_data)

    # Add omitted flash info:
    try:
        omitted_flash_frame_log = data["items"]["behavior"]["omitted_flash_frame_log"]
    except KeyError:
        # For sessions for which there were no omitted flashes
        omitted_flash_frame_log = dict()

    omitted_flash_list = []
    for _, omitted_flash_frames in omitted_flash_frame_log.items():
        stim_frames = visual_stimuli_df["frame"].values
        omitted_flash_frames = np.array(omitted_flash_frames)

        # Test offsets of omitted flash frames
        # to see if they are in the stim log
        offsets = np.arange(-3, 4)
        offset_arr = np.add(
            np.repeat(omitted_flash_frames[:, np.newaxis], offsets.shape[0], axis=1),
            offsets,
        )
        matched_any_offset = np.any(np.isin(offset_arr, stim_frames), axis=1)

        #  Remove omitted flashes that also exist in the stimulus log
        was_true_omitted = np.logical_not(matched_any_offset)  # bool
        omitted_flash_frames_to_keep = omitted_flash_frames[was_true_omitted]

        # Have to remove frames that are double-counted in omitted log
        omitted_flash_list += list(np.unique(omitted_flash_frames_to_keep))

    omitted = np.ones_like(omitted_flash_list).astype(bool)
    time = [time[fi] for fi in omitted_flash_list]
    omitted_df = pd.DataFrame(
        {
            "omitted": omitted,
            "frame": omitted_flash_list,
            "time": time,
            "image_name": "omitted",
        }
    )

    df = pd.concat((visual_stimuli_df, omitted_df), sort=False).sort_values("frame").reset_index()
    return df


def get_image_names(behavior_stimulus_file) -> Set[str]:
    """Gets set of image names shown during behavior session"""
    stimuli = behavior_stimulus_file["stimuli"]
    image_names = set()
    for stim_dict in stimuli.values():
        for attr_name, attr_value, _, _ in stim_dict["set_log"]:
            if attr_name.lower() == "image":
                image_names.add(attr_value)
    return image_names


def is_change_event(stimulus_presentations: pd.DataFrame) -> pd.Series:
    """
    Returns whether a stimulus is a change stimulus
    A change stimulus is defined as the first presentation of a new image_name
    Omitted stimuli are ignored
    The first stimulus in the session is ignored

    :param stimulus_presentations
        The stimulus presentations table

    :return: is_change: pd.Series indicating whether a given stimulus is a
        change stimulus
    """
    stimuli = stimulus_presentations["image_name"]

    # exclude omitted stimuli
    stimuli = stimuli[~stimulus_presentations["omitted"]]

    prev_stimuli = stimuli.shift()

    # exclude first stimulus
    stimuli = stimuli.iloc[1:]
    prev_stimuli = prev_stimuli.iloc[1:]

    is_change = stimuli != prev_stimuli

    stimulus_presentations = stimulus_presentations[~stimulus_presentations.index.duplicated(keep="first")]
    is_change = is_change[~is_change.index.duplicated(keep="first")]

    # reset back to original index
    is_change = is_change.reindex(stimulus_presentations.index).rename("is_change")

    # Excluded stimuli are not change events; reindex may have introduced NaN
    # which upcasts the bool series to object — cast back explicitly.
    is_change = is_change.fillna(False).astype(bool)

    return is_change


def get_flashes_since_change(
    stimulus_presentations: pd.DataFrame,
) -> pd.Series:
    """Calculate the number of times an images is flashed between changes.

    Parameters
    ----------
    stimulus_presentations : pandas.DataFrame
        Table of presented stimuli with ``is_change`` column already
        calculated.

    Returns
    -------
    flashes_since_change : pandas.Series
        Number of times the same image is flashed between image changes.
    """
    flashes_since_change = pd.Series(
        data=np.zeros(len(stimulus_presentations), dtype=float),
        index=stimulus_presentations.index,
        name="flashes_since_change",
        dtype="int",
    )
    for idx, (pd_index, row) in enumerate(stimulus_presentations.iterrows()):
        omitted = row["omitted"]
        if pd.isna(row["omitted"]):
            omitted = False
        if row["image_name"] == "omitted" or omitted:
            flashes_since_change.iloc[idx] = flashes_since_change.iloc[idx - 1]
        else:
            if row["is_change"] or idx == 0:
                flashes_since_change.iloc[idx] = 0
            else:
                flashes_since_change.iloc[idx] = flashes_since_change.iloc[idx - 1] + 1
    return flashes_since_change


def add_active_flag(stim_pres_table: pd.DataFrame, trials: pd.DataFrame) -> pd.DataFrame:
    """Mark the active stimuli by lining up the stimulus times with the
    trials times.

    Parameters
    ----------
    stim_pres_table : pandas.DataFrame
        Stimulus table to add active column to.
    trials : pandas.DataFrame
        Trials table to align with the stimulus table.

    Returns
    -------
    stimulus_table : pandas.DataFrame
        Copy of ``stim_pres_table`` with added acive column.
    """
    if "active" in stim_pres_table.columns:
        return stim_pres_table
    else:
        active = pd.Series(
            data=np.zeros(len(stim_pres_table), dtype=bool),
            index=stim_pres_table.index,
            name="active",
        )
        stim_mask = (
            (stim_pres_table.start_time > trials.start_time.min())
            & (stim_pres_table.start_time < trials.stop_time.max())
            & (~stim_pres_table.image_name.isna())
        )
        active[stim_mask] = True

        # Clean up potential stimuli that fall outside in time of the trials
        # but are part of the "active" stimulus block.
        if "stimulus_block" in stim_pres_table.columns:
            for stim_block in stim_pres_table["stimulus_block"].unique():
                block_mask = stim_pres_table["stimulus_block"] == stim_block
                if np.any(active[block_mask]):
                    active[block_mask] = True
        stim_pres_table["active"] = active
        return stim_pres_table


def compute_trials_id_for_stimulus(stim_pres_table: pd.DataFrame, trials_table: pd.DataFrame) -> pd.Series:
    """Add an id to allow for merging of the stimulus presentations
    table with the trials table.

    If stimulus_block is not available as a column in the input table, return
    an empty set of trials_ids.

    Parameters
    ----------
    stim_pres_table : pandas.DataFrame
        Pandas stimulus table to create trials_id from.
    trials_table : pandas.DataFrame
        Trials table to create id from using trial start times.

    Returns
    -------
    trials_ids : pd.Series
        Unique id to allow merging of the stim table with the trials table.
        Null values are represented by -1.

    Note
    ----
    ``trials_id`` values are copied from active stimulus blocks into
    passive stimulus/replay blocks that contain the same image ordering and
    length.
    """
    # Create a placeholder for the trials_id.
    trials_ids = pd.Series(
        data=np.full(len(stim_pres_table), INT_NULL, dtype=int),
        index=stim_pres_table.index,
        name="trials_id",
    ).astype("int")

    # Find stimulus blocks that start within a trial. Copy the trial_id
    # into our new trials_ids series. For some sessions there are gaps in
    # between one trial's end and the next's stop time so we account for this
    # by only using the max time for all trials as the limit.
    max_trials_stop = trials_table.stop_time.max()
    for idx, trial in trials_table.iterrows():
        stim_mask = (
            (stim_pres_table.start_time > trial.start_time)
            & (stim_pres_table.start_time < max_trials_stop)
            & (~stim_pres_table.image_name.isna())
        )
        trials_ids[stim_mask] = idx

    # Return input frame if the stimulus_block or active is not available.
    if "stimulus_block" not in stim_pres_table.columns or "active" not in stim_pres_table.columns:
        return trials_ids
    active_sorted = stim_pres_table.active

    # The code below finds all stimulus blocks that contain images/trials
    # and attempts to detect blocks that are identical to copy the associated
    # trials_ids into those blocks. In the parlance of the data this is
    # copying the active stimulus block data into the passive stimulus block.

    # Get the block ids for the behavior trial presentations
    stim_blocks = stim_pres_table.stimulus_block
    stim_image_names = stim_pres_table.image_name
    active_stim_blocks = stim_blocks[active_sorted].unique()
    # Find passive blocks that show images for potential copying of the active
    # into a passive stimulus block.
    passive_stim_blocks = stim_blocks[np.logical_and(~active_sorted, ~stim_image_names.isna())].unique()

    # Copy the trials_id into the passive block if it exists.
    if len(passive_stim_blocks) > 0:
        for active_stim_block in active_stim_blocks:
            active_block_mask = stim_blocks == active_stim_block
            active_images = stim_image_names[active_block_mask].values
            for passive_stim_block in passive_stim_blocks:
                passive_block_mask = stim_blocks == passive_stim_block
                if np.array_equal(active_images, stim_image_names[passive_block_mask].values):
                    trials_ids.loc[passive_block_mask] = trials_ids[active_block_mask].values

    return trials_ids.sort_index()


def fix_omitted_end_frame(stim_pres_table: pd.DataFrame) -> pd.DataFrame:
    """Fill NaN ``end_frame`` values for omitted frames.

    Additionally, change type of ``end_frame`` to int.

    Parameters
    ----------
    stim_pres_table : `pandas.DataFrame`
        Input stimulus table to fix/fill omitted ``end_frame`` values.

    Returns
    -------
    output : `pandas.DataFrame`
        Copy of input DataFrame with filled omitted, ``end_frame`` values and
        fixed typing.
    """
    median_stim_frame_duration = np.nanmedian(stim_pres_table["end_frame"] - stim_pres_table["start_frame"])
    omitted_end_frames = stim_pres_table[stim_pres_table["omitted"]]["start_frame"] + median_stim_frame_duration
    stim_pres_table.loc[stim_pres_table["omitted"], "end_frame"] = omitted_end_frames

    stim_dtypes = stim_pres_table.dtypes.to_dict()
    stim_dtypes["start_frame"] = int
    stim_dtypes["end_frame"] = int

    return stim_pres_table.astype(stim_dtypes)


def compute_is_sham_change(stim_df: pd.DataFrame, trials: pd.DataFrame) -> pd.DataFrame:
    """Add is_sham_change to stimulus presentation table.

    Parameters
    ----------
    stim_df : pandas.DataFrame
        Stimulus presentations table to add is_sham_change to.
    trials : pandas.DataFrame
        Trials data frame to pull info from to create

    Returns
    -------
    stimulus_presentations : pandas.DataFrame
        Input ``stim_df`` DataFrame with the is_sham_change column added.
    """
    if "trials_id" not in stim_df.columns or "active" not in stim_df.columns or "stimulus_block" not in stim_df.columns:
        return stim_df
    stim_trials = stim_df.merge(trials, left_on="trials_id", right_index=True, how="left")
    catch_frames = stim_trials[stim_trials["catch"].fillna(False)]["change_frame"].unique()

    stim_df["is_sham_change"] = False
    catch_flashes = stim_df[stim_df["start_frame"].isin(catch_frames)].index.values
    stim_df.loc[catch_flashes, "is_sham_change"] = True

    stim_blocks = stim_df.stimulus_block
    stim_image_names = stim_df.image_name
    active_stim_blocks = stim_blocks[stim_df.active].unique()
    # Find passive blocks that show images for potential copying of the active
    # into a passive stimulus block.
    passive_stim_blocks = stim_blocks[np.logical_and(~stim_df.active, ~stim_image_names.isna())].unique()

    # Copy the trials_id into the passive block if it exists.
    if len(passive_stim_blocks) > 0:
        for active_stim_block in active_stim_blocks:
            active_block_mask = stim_blocks == active_stim_block
            active_images = stim_image_names[active_block_mask].values
            for passive_stim_block in passive_stim_blocks:
                passive_block_mask = stim_blocks == passive_stim_block
                if np.array_equal(active_images, stim_image_names[passive_block_mask].values):
                    stim_df.loc[passive_block_mask, "is_sham_change"] = stim_df[active_block_mask][
                        "is_sham_change"
                    ].values

    return stim_df.sort_index()


def fingerprint_from_stimulus_file(
    stimulus_presentations: pd.DataFrame,
    stimulus_file,
    stimulus_timestamps,
    fingerprint_name,
):
    """
    Instantiates `fingerprintStimulus` from stimulus file

    Parameters
    ----------
    stimulus_presentations:
        Table containing previous stimuli
    stimulus_file
        BehaviorStimulusFile
    stimulus_timestamps
        StimulusTimestamps

    Returns
    -------
    `fingerprintStimulus`
        Instantiated fingerprintStimulus
    """
    fingerprint_stim = stimulus_file["items"]["behavior"]["items"][fingerprint_name]["static_stimulus"]

    # not sure why this was here
    # n_repeats = fingerprint_stim["runs"]

    # spontaneous + fingerprint indices relative to start of session
    stimulus_session_frame_indices = np.array(
        stimulus_file["items"]["behavior"]["items"][fingerprint_name]["frame_indices"]
    )

    # not sure why this was here
    # movie_length = int(len(fingerprint_stim["sweep_frames"]) / n_repeats)

    # Start index within the spontaneous + fingerprint block
    movie_start_index = sum(1 for frame in fingerprint_stim["frame_list"] if frame == -1)
    res = []

    sweep_frames = fingerprint_stim["sweep_frames"]
    sweep_table = [fingerprint_stim["sweep_table"][i] for i in fingerprint_stim["sweep_order"]]
    dimnames = fingerprint_stim["dimnames"]

    res = []

    for i, stimulus_frame_indices in enumerate(sweep_frames):
        stimulus_frame_indices = np.array(stimulus_frame_indices).astype(int)

        # Adjust for spontaneous block offset
        valid_indices = np.clip(
            stimulus_frame_indices + movie_start_index,
            0,
            len(stimulus_session_frame_indices) - 1,
        )

        start_frame, end_frame = stimulus_session_frame_indices[valid_indices]
        start_time = stimulus_timestamps[start_frame]
        stop_time = stimulus_timestamps[min(end_frame + 1, len(stimulus_timestamps) - 1)]

        stim_row = sweep_table[i]
        stim_info = dict(zip(dimnames, stim_row))

        res.append(
            {
                "start_time": start_time,
                "stop_time": stop_time,
                "start_frame": start_frame,
                "end_frame": end_frame,
                "duration": stop_time - start_time,
                **stim_info,  # unpack stimulus parameters
            }
        )

    table = pd.DataFrame(res)

    # Add static columns
    table["stim_block"] = stimulus_presentations["stim_block"].max() + 2
    table["stim_name"] = fingerprint_name

    # Coerce ints cleanly
    table = table.astype({c: "int64" for c in table.select_dtypes(include="int")})

    return table


def clean_position_and_contrast(df):
    """Clean position and contrast columns in stimulus presentation table.

    Parameters
    ----------
    df : pandas.DataFrame
        Input stimulus presentation table to clean.

    Returns
    -------
    df : pandas.DataFrame
        Cleaned stimulus presentation table with position and contrast columns
        processed.
    """

    def safe_split_pos(x):
        """Safely split position into x and y coordinates."""

        if isinstance(x, (list, tuple)) and len(x) == 2:
            return pd.Series([float(x[0]), float(x[1])])
        else:
            return pd.Series([np.nan, np.nan])

    if "Pos" in df.columns:
        df[["pos_x", "pos_y"]] = df["Pos"].apply(safe_split_pos)
        df.drop(columns=["Pos"], inplace=True)

    if "contrast" in df.columns:
        df["contrast"] = df["contrast"].apply(lambda x: x[0] if isinstance(x, list) else x).astype(float)
    return df


def _load_and_validate_stimulus_presentations(data, stimulus_timestamps):
    """
    Load and validate stimulus presentations from stimulus file data.

    Parameters
    ----------
    data : dict
        The loaded stimulus file data.
    stimulus_timestamps : StimulusTimestamps
        Timestamps of the stimuli.

    Returns
    -------
    pd.DataFrame
        Cleaned and validated stimulus presentations dataframe.
    """
    raw_stim_pres_df = get_stimulus_presentations(data, stimulus_timestamps)
    raw_stim_pres_df = raw_stim_pres_df.drop(columns=["index"])
    raw_stim_pres_df = check_for_errant_omitted_stimulus(input_df=raw_stim_pres_df)
    return raw_stim_pres_df


def from_stimulus_file(
    stimulus_file,
    stimulus_timestamps,
    limit_to_images: Optional[List] = None,
    column_list: Optional[List[str]] = None,
    fill_omitted_values: bool = True,
    project_code: Optional[str] = None,
):
    """Get stimulus presentation data.

    Parameters
    ----------
    stimulus_file : BehaviorStimulusFile
        Input stimulus_file to create presentations dataframe from.
    stimulus_timestamps : StimulusTimestamps
        Timestamps of the stimuli
    behavior_session_id : int
        LIMS id of behavior session
    trials: Trials
        Object to create trials_id column in Presentations table
        allowing for mering of the two tables.
    limit_to_images : Optional, list of str
        Only return images given by these image names
    column_list : Optional, list of str
        The columns and order of columns in the final dataframe
    fill_omitted_values : Optional, bool
        Whether to fill stop_time and duration for omitted frames
    project_code: Optional, ProjectCode
        For released datasets, provide a project code
        to produce explicitly named stimulus_block column values in the
        column stimulus_block_name

    Returns
    -------
    output_presentations: Presentations
        Object with a table whose rows are stimulus presentations
        (i.e. a given image, for a given duration, typically 250 ms)
        and whose columns are presentation characteristics.
    """
    data = pkl.load_pkl(stimulus_file)
    raw_stim_pres_df = _load_and_validate_stimulus_presentations(data, stimulus_timestamps)

    # print(raw_stim_pres_df.head(100))
    # Fill in nulls for image_name
    # This makes two assumptions:
    #   1. Nulls in `image_name` should be "gratings_<orientation>"
    #   2. Gratings are only present (or need to be fixed) when all
    #      values for `image_name` are null.
    if pd.isnull(raw_stim_pres_df["image_name"]).all():
        if not pd.isnull(raw_stim_pres_df["orientation"]).all():
            raw_stim_pres_df["image_name"] = raw_stim_pres_df["orientation"].apply(lambda x: f"gratings_{x}")
        else:
            raise ValueError("All values for 'orientation' and " "'image_name are null.")
    # print(raw_stim_pres_df.head(100))
    stimulus_metadata_df = get_stimulus_metadata(data)
    # print("metadata")
    # print(stimulus_metadata_df)

    idx_name = raw_stim_pres_df.index.name
    if idx_name is None:
        return raw_stim_pres_df
    raw_stim_pres_df = raw_stim_pres_df.drop(columns=["orientation"])
    # print(raw_stim_pres_df.head(100))
    stimulus_index_df = (
        raw_stim_pres_df.reset_index()
        .merge(stimulus_metadata_df.reset_index(), on=["image_name"], how="outer")
        .set_index(idx_name)
    )
    # print("stimulus_index_df")
    # print(stimulus_index_df.head(100))

    # Only select columns that actually exist in the DataFrame
    required_columns = [
        "image_set",
        "image_index",
        "start_time",
        "phase",
        "size",
        "orientation",
        "spatial_frequency",
    ]
    available_columns = [col for col in required_columns if col in stimulus_index_df.columns]

    stimulus_index_df = (
        stimulus_index_df[available_columns]
        .rename(columns={"start_time": "timestamps"})
        .sort_index()
        .set_index("timestamps", drop=True)
    )
    if not stimulus_index_df["image_index"].isna().any():
        stimulus_index_df["image_index"] = stimulus_index_df["image_index"].astype(int)
    stim_pres_df = raw_stim_pres_df.merge(
        stimulus_index_df,
        left_on="start_time",
        right_index=True,
        how="left",
    )

    # Sort columns then drop columns which contain only all NaN values
    stim_pres_df = stim_pres_df[sorted(stim_pres_df)].dropna(axis=1, how="all")
    if limit_to_images is not None:
        stim_pres_df = stim_pres_df[stim_pres_df["image_name"].isin(limit_to_images)]
        stim_pres_df.index = pd.Index(range(stim_pres_df.shape[0]), name=stim_pres_df.index.name)

    stim_pres_df["stim_block"] = 0
    stim_pres_df["stim_name"] = get_stimulus_name(data)

    stim_pres_df = fix_omitted_end_frame(stim_pres_df)

    for fingerprint_name, stim_data in data["items"]["behavior"]["items"].items():
        if "static_stimulus" in stim_data:
            stim_pres_df = add_fingerprint_stimulus(
                stimulus_presentations=stim_pres_df,
                stimulus_file=data,
                stimulus_timestamps=stimulus_timestamps,
                fingerprint_name=fingerprint_name,
            )
    stim_pres_df = postprocess(
        presentations=stim_pres_df,
        fill_omitted_values=fill_omitted_values,
        coerce_bool_to_boolean=True,
    )
    df = stim_pres_df

    # Identify duplicates based on "start_time" and "image_name"
    dupes = df[df.duplicated(subset=["start_time", "image_name"], keep=False)]

    # Make sure image_set exists in dupes:
    if "image_set" in df.columns:
        # Identify duplicated image presentations
        dupes = df[df.duplicated(subset=["start_time", "image_name"], keep=False)]

        # Remove 'pkl' rows when there's a matching non-pkl row
        pkl_mask = dupes["image_set"].str.contains("pkl", na=False)
        non_pkl_dupes = dupes[~pkl_mask][["start_time", "image_name"]]
        duplicate_keys = set(non_pkl_dupes.itertuples(index=False, name=None))

        df = df[
            ~df.set_index(["start_time", "image_name"]).index.isin(duplicate_keys)
            | ~df["image_set"].str.contains("pkl", na=False)
        ]

        # Normalize image_set by stripping `.pkl` extension if present
        df["image_set"] = df["image_set"].where(
            ~df["image_set"].fillna("").str.endswith(".pkl"),
            df["image_set"].str.extract(r"([^/]+)\.pkl$")[0],
        )

    # Reset index for clarity
    df.reset_index(drop=True, inplace=True)
    df = names.map_column_names(df, constants.default_column_renames)
    duplicates = df.columns[df.columns.duplicated()].unique()
    # Merge each group of duplicated columns
    for col in duplicates:
        # Select all duplicate columns with this name
        dup_cols = df.loc[:, df.columns == col]

        # Merge: pick the first non-null value in the row
        merged = dup_cols.bfill(axis=1).iloc[:, 0]

        # Drop all the duplicate columns
        df = df.drop(columns=dup_cols.columns)

        # Add back the merged column
        df[col] = merged
    df = remove_short_sandwiched_spontaneous(df)
    df = clean_position_and_contrast(df)

    df.drop(columns=["stim_block"], inplace=True, errors="ignore")
    df = df.drop(columns=["start_frame", "end_frame"], errors="ignore")

    return (df, column_list)


def get_is_image_novel(
    image_names: List[str],
    behavior_session_id: int,
) -> Dict[str, bool]:
    """
    Returns whether each image in `image_names` is novel for the mouse

    Parameters
    ----------
    image_names:
        List of image names
    behavior_session_id
        LIMS behavior session id

    Returns
    -------
    Dict mapping image name to is_novel
    """

    # TODO: FIND A WAY TO DO THIS WITHOUT LIMS?

    return False
    """
    mouse = Mouse.from_behavior_session_id(
        behavior_session_id=behavior_session_id
    )
    prior_images_shown = mouse.get_images_shown(
        up_to_behavior_session_id=behavior_session_id
    )

    image_names = set(
        [x for x in image_names if x != "omitted" and type(x) is str]
    )
    is_novel = {
        f"{image_name}": image_name not in prior_images_shown
        for image_name in image_names
    }
    return is_novel
    """


def postprocess(
    presentations: pd.DataFrame,
    fill_omitted_values=True,
    coerce_bool_to_boolean=True,
    omitted_time_duration: float = 0.25,
) -> pd.DataFrame:
    """
    Applies further processing to `presentations`

    Parameters
    ----------
    presentations
        Presentations df
    fill_omitted_values
        Whether to fill stop time and duration for omitted flashes
    coerce_bool_to_boolean
        Whether to coerce columns of "Object" dtype that are truly bool
        to nullable "boolean" dtype
    omitted_time_duration
        Amount of time a stimuli is omitted for in seconds"""
    df = presentations
    if fill_omitted_values:
        fill_missing_values_for_omitted_flashes(df=df, omitted_time_duration=omitted_time_duration)
    if coerce_bool_to_boolean:
        bool_like_cols = {}
        for c in df.select_dtypes("O").columns:
            non_na_values = df[c][~df[c].isna()]
            # Skip columns with list-like elements
            if non_na_values.apply(lambda x: isinstance(x, list)).any():
                continue
            unique_vals = set(non_na_values.unique())
            if unique_vals.issubset({True, False}):
                bool_like_cols[c] = "boolean"
        df = df.astype(bool_like_cols)

    df = check_for_errant_omitted_stimulus(input_df=df)
    return df


def check_for_errant_omitted_stimulus(
    input_df: pd.DataFrame,
) -> pd.DataFrame:
    """Check if the first entry in the DataFrame is an omitted stimulus.

    This shouldn't happen and likely reflects some sort of camstim error
    with appending frames to the omitted flash frame log. See
    explanation here:
    https://github.com/AllenInstitute/AllenSDK/issues/2577

    Parameters
    ----------/
    input_df : DataFrame
        Input stimulus table to check for "omitted" stimulus.

    Returns
    -------
    modified_df : DataFrame
        Dataframe with omitted stimulus removed from first row or if not
        found, return input_df unmodified.
    """

    def safe_omitted_check(input_df: pd.Series, stimulus_block: Optional[int]):
        """
        Check if the first row in the input_df is an omitted stimulus.

        Parameters
        ----------
        input_df : pd.Series
            Input stimulus table to check for "omitted" stimulus.
        stimulus_block : Optional[int]
            Stimulus block to check for omitted stimulus in.

        Returns
        -------
        input_df : pd.Series
            Dataframe with omitted stimulus removed from first row or if not
        """
        if stimulus_block is not None:
            first_row = input_df[input_df["stimulus_block"] == stim_block].iloc[0]
        else:
            first_row = input_df.iloc[0]

        if not pd.isna(first_row["omitted"]):
            if first_row["omitted"]:
                input_df = input_df.drop(first_row.name, axis=0)
        return input_df

    if "omitted" in input_df.columns and len(input_df) > 0:
        if "stimulus_block" in input_df.columns:
            for stim_block in input_df["stimulus_block"].unique():
                input_df = safe_omitted_check(input_df=input_df, stimulus_block=stim_block)
        else:
            input_df = safe_omitted_check(input_df=input_df, stimulus_block=None)
    return input_df


def fill_missing_values_for_omitted_flashes(df: pd.DataFrame, omitted_time_duration: float = 0.25) -> pd.DataFrame:
    """
    This function sets the stop time for a row that is an omitted
    stimulus. An omitted stimulus is a stimulus where a mouse is
    shown only a grey screen and these last for 250 milliseconds.
    These do not include a stop_time or end_frame like other stimuli in
    the stimulus table due to design choices.

    Parameters
    ----------
    df
        Stimuli presentations dataframe
    omitted_time_duration
        Amount of time a stimulus is omitted for in seconds
    """
    omitted = df["omitted"].fillna(False)
    df.loc[omitted, "stop_time"] = df.loc[omitted, "start_time"] + omitted_time_duration
    df.loc[omitted, "duration"] = omitted_time_duration
    return df


def get_spontaneous_stimulus(
    stimulus_presentations_table: pd.DataFrame,
) -> pd.DataFrame:
    """The spontaneous stimulus is a gray screen shown in between
    different stimulus blocks. This method finds any gaps in the stimulus
    presentations. These gaps are assumed to be spontaneous stimulus.

    Parameters
    ---------
    stimulus_presentations_table : pd.DataFrame
        Input stimulus presentations table.

    Returns
    -------
    output_frame : pd.DataFrame
        stimulus_presentations_table with added spotaneous stimulus blocks
        added.

    Raises
    ------
    RuntimeError if there are any gaps in stimulus blocks > 1
    """
    res = []
    # Check for 5 minute gray screen stimulus block at the start of the
    # movie. We give some leeway around 5 minutes at 285 seconds to account
    # for some sessions which have start times slightly less than 300
    # seconds. This also makes sure that presentations that start slightly
    # late are not erroneously added as a "grey screen".
    if (
        stimulus_presentations_table.iloc[0]["start_frame"] > 0
        and stimulus_presentations_table.iloc[0]["start_time"] > 285
    ):
        res.append(
            {
                "duration": stimulus_presentations_table.iloc[0]["start_time"],
                "start_time": 0,
                "stop_time": stimulus_presentations_table.iloc[0]["start_time"],
                "start_frame": 0,
                "end_frame": stimulus_presentations_table.iloc[0]["start_frame"],
                "stim_block": 0,
                "stim_name": "spontaneous",
            }
        )
        # Increment the stimulus blocks by 1 to to account for the
        # new stimulus at the start of the file.
        stimulus_presentations_table["stim_block"] += 1

    spontaneous_stimulus_blocks = get_spontaneous_block_indices(
        stimulus_blocks=(stimulus_presentations_table["stim_block"].values)
    )

    for spontaneous_block in spontaneous_stimulus_blocks:
        prev_stop_time = stimulus_presentations_table[
            stimulus_presentations_table["stim_block"] == spontaneous_block - 1
        ]["stop_time"].max()
        prev_end_frame = stimulus_presentations_table[
            stimulus_presentations_table["stim_block"] == spontaneous_block - 1
        ]["end_frame"].max()
        next_start_time = stimulus_presentations_table[
            stimulus_presentations_table["stim_block"] == spontaneous_block + 1
        ]["start_time"].min()
        next_start_frame = stimulus_presentations_table[
            stimulus_presentations_table["stim_block"] == spontaneous_block + 1
        ]["start_frame"].min()
        res.append(
            {
                "duration": next_start_time - prev_stop_time,
                "start_time": prev_stop_time,
                "stop_time": next_start_time,
                "start_frame": prev_end_frame,
                "end_frame": next_start_frame,
                "stim_block": spontaneous_block,
                "stim_name": "spontaneous",
            }
        )

    res = pd.DataFrame(res)

    return pd.concat([stimulus_presentations_table, res]).sort_values("start_frame")


def make_spontaneous_stimulus(
    stimulus_presentations_table: pd.DataFrame,
) -> pd.DataFrame:
    """Identifies spontaneous stimulus
    (gray screen shown in between stimulus blocks)
    by detecting gaps in stimulus presentations.

    Parameters
    ----------
    stimulus_presentations_table : pd.DataFrame
        Input stimulus presentations table.

    Returns
    -------
    pd.DataFrame
        The stimulus presentations table with added spontaneous stimulus
        blocks.
    """

    # Ensure the table is sorted by start time
    stimulus_presentations_table = stimulus_presentations_table.sort_values("start_time").reset_index(drop=True)

    # Identify gaps between consecutive stimulus presentations
    gaps = (
        stimulus_presentations_table["start_time"].iloc[1:].values
        - stimulus_presentations_table["stop_time"].iloc[:-1].values
    )
    gap_indices = np.where(gaps > 0)[0]  # Indices where gaps exist

    spontaneous_entries = []

    for idx in gap_indices:
        prev_row = stimulus_presentations_table.iloc[idx]
        next_row = stimulus_presentations_table.iloc[idx + 1]

        spontaneous_entries.append(
            {
                "start_time": prev_row["stop_time"],
                "stop_time": next_row["start_time"],
                "duration": next_row["start_time"] - prev_row["stop_time"],
                "start_frame": prev_row["end_frame"],
                "end_frame": next_row["start_frame"],
                "stim_block": prev_row["stim_block"] + 1,  # Assign new block number
                "stim_name": "spontaneous",
            }
        )

    # Convert spontaneous entries into a DataFrame
    spontaneous_df = pd.DataFrame(spontaneous_entries)

    if spontaneous_df.empty:
        # No gaps found, return original table
        return stimulus_presentations_table

    # Adjust the stim_block values in the original table to account for
    # inserted spontaneous blocks
    stimulus_presentations_table["stim_block"] += (
        stimulus_presentations_table["stim_block"] >= spontaneous_df["stim_block"].min()
    ).astype(int)

    # Merge and return sorted table
    return pd.concat([stimulus_presentations_table, spontaneous_df]).sort_values("start_frame").reset_index(drop=True)


def add_fingerprint_stimulus(
    stimulus_presentations: pd.DataFrame,
    stimulus_file,
    stimulus_timestamps,
    fingerprint_name,
) -> pd.DataFrame:
    """Adds the fingerprint stimulus and the preceding gray screen to
    the stimulus presentations table

    Returns
    -------
    pd.DataFrame: stimulus presentations with gray screen + fingerprint
    movie added"""

    fingerprint_stimulus = fingerprint_from_stimulus_file(
        stimulus_presentations=stimulus_presentations,
        stimulus_file=stimulus_file,
        stimulus_timestamps=stimulus_timestamps,
        fingerprint_name=fingerprint_name,
    )

    stimulus_presentations = pd.concat([stimulus_presentations, fingerprint_stimulus])
    # stimulus_presentations = get_spontaneous_stimulus(
    #    stimulus_presentations_table=stimulus_presentations
    # )
    # reset index to go from 0...end
    stimulus_presentations = make_spontaneous_stimulus(stimulus_presentations)
    stimulus_presentations.index = pd.Index(
        np.arange(0, stimulus_presentations.shape[0]),
        name=stimulus_presentations.index.name,
        dtype=stimulus_presentations.index.dtype,
    )
    return stimulus_presentations


def get_spontaneous_block_indices(stimulus_blocks: np.ndarray) -> np.ndarray:
    """Gets the indices where there is a gap in stimulus block. This is
    where spontaneous blocks are.
    Example: stimulus blocks are [0, 2, 3]. There is a spontaneous block at 1.

    Parameters
    ----------
    stimulus_blocks: Stimulus blocks in the stimulus presentations table

    Notes
    -----
    This doesn't support a spontaneous block appearing at the beginning or
    end of a session

    Returns
    -------
    np.array: spontaneous stimulus blocks
    """
    blocks = np.sort(np.unique(stimulus_blocks))
    block_diffs = np.diff(blocks)
    if (block_diffs > 2).any():
        raise RuntimeError(
            f"There should not be any stimulus block " f"diffs greater than 2. The stimulus " f"blocks are {blocks}"
        )

    # i.e. if the current blocks are [0, 2], then block_diffs will
    # be [2], with a gap (== 2) at index 0, meaning that the spontaneous block
    # is at index 1
    block_indices = blocks[np.where(block_diffs == 2)[0]] + 1
    return block_indices


def get_stimulus_name(stim_file) -> str:
    """
    Get the image stimulus name by parsing the file path of the image set.

    If no image set, check for gratings and return "behavior" if not found.

    Parameters
    ----------
    stimulus_file : BehaviorStimulusFile
        Stimulus pickle file to parse.

    Returns
    -------
    stimulus_name : str
        Name of the image stimulus from the image file path set shown to
        the mouse.
    """
    try:
        stimulus_name = Path(stim_file["items"]["behavior"]["images"]["image_set"]).stem.split(".")[0]
    except KeyError:
        try:
            stimulus_name = Path(stim_file["items"]["behavior"]["stimuli"]["images"]["image_set"]).stem.split(".")[0]
        except KeyError:
            print("No image set found in stimulus file")
            stimulus_name = "behavior"
        # if we can't find the images key in the stimuli, check for the
        # name ``grating`` as the stimulus. If not add generic
        # ``behavior``.
        if "grating" in stim_file["items"]["behavior"]["stimuli"].keys():
            stimulus_name = "grating"
        else:
            stimulus_name = "behavior"
    return stimulus_name



"""Data models of extracted information from tiff files."""

from enum import Enum
from pathlib import Path
from typing import List

from pydantic import BaseModel


class TifFileGroup(str, Enum):
    """Type of stimulation a group of files belongs to"""

    BEHAVIOR = "behavior"
    PHOTOSTIM = "photostim"
    SPONTANEOUS = "spontaneous"
    STACK = "stack"


class RawImageInfo(BaseModel):
    """Raw metadata from a tif file"""

    reader_metadata_header: dict
    reader_metadata_json: dict
    # The reader descriptions for the last tif file
    reader_descriptions: List[dict]
    # Looks like [620, 800, 800]
    # [num_of_frames, pixel_width, pixel_height]?
    reader_shape: List[int]


class ExtractedInfoItem(BaseModel):
    """ExtractedInfo that can be used to build Session metadata."""

    raw_info_first_file: RawImageInfo
    raw_info_last_file: RawImageInfo
    tif_file_group: TifFileGroup
    file_stem: str
    files: List[Path]


class ExtractedInfo(BaseModel):
    """Main model to be used downstream."""

    info: List[ExtractedInfoItem]



"""SmartSPIM job settings configuration."""

from pathlib import Path
from typing import Literal, Optional, Union

from pydantic import Field
from aind_metadata_extractor.core import BaseJobSettings


class JobSettings(BaseJobSettings):
    """Data to be entered by the user."""

    # Field can be used to switch between different acquisition etl jobs
    job_settings_name: Literal["SmartSPIM"] = "SmartSPIM"
    raw_dataset_path: Optional[Union[Path, str]] = Field(
        default=None, description=("Deprecated, use input_source instead.")
    )
    subject_id: str
    acquisition_type: str

    # Metadata names
    asi_filename: str = Field(
        default="derivatives/ASI_logging.txt",
        description="Path to ASI logging file.",
    )
    mdata_filename_json: str = Field(
        default="derivatives/metadata.json",
        description=("Path to metadata file, expected to be a .json or .txt file."),
    )
    # Fetch info provided by microscope operators in SLIMS
    processing_manifest_path: Optional[Union[Path, str]] = Field(
        default="derivatives/processing_manifest.json",
        description=("Deprecated, use metadata_service_path instead."),
    )
    metadata_service_path: str



"""SmartSPIM extractor model"""

from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import BaseModel


class FileMetadataModel(BaseModel):
    """Model for metadata extracted from microscope files"""

    session_config: Dict[str, Any]
    wavelength_config: Dict[str, Any]
    tile_config: Dict[str, Any]
    session_start_time: datetime
    session_end_time: datetime
    filter_mapping: Dict[str, Any]


class SlimsMetadataModel(BaseModel):
    """Model for metadata extracted from SLIMS API response"""

    experiment_run_created_on: Optional[str] = None
    order_created_by: Optional[str] = None
    order_project_id: Optional[str] = None
    specimen_id: Optional[str] = None
    subject_id: Optional[str] = None
    protocol_name: Optional[str] = None
    protocol_id: Optional[str] = None
    date_performed: Optional[str] = None
    chamber_immersion_medium: Optional[str] = None
    sample_immersion_medium: Optional[str] = None
    chamber_refractive_index: Optional[str] = None
    sample_refractive_index: Optional[str] = None
    instrument_id: Optional[str] = None
    experimenter_name: Optional[str] = None
    z_direction: Optional[str] = None
    y_direction: Optional[str] = None
    x_direction: Optional[str] = None
    imaging_channels: Optional[List[str]] = None
    stitching_channels: Optional[str] = None
    ccf_registration_channels: Optional[str] = None
    cell_segmentation_channels: Optional[str] = None


class SmartspimModel(BaseModel):
    """SmartSPIM extractor model for intermediate data structure"""

    acquisition_type: str
    file_metadata: FileMetadataModel
    slims_metadata: SlimsMetadataModel



"""Utilities for working with stimulus data."""

import ast
import functools
import logging
import re
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

import aind_metadata_extractor.utils.camstim_sync.pkl_utils as pkl
import aind_metadata_extractor.utils.camstim_sync.sync_utils as sync

logger = logging.getLogger(__name__)

DROP_PARAMS = (  # psychopy boilerplate, more or less
    "autoLog",
    "autoDraw",
    "win",
)

REPR_PARAMS_RE = re.compile(r"([a-z0-9]+=[^=]+)[,\)]", re.IGNORECASE)
REPR_CLASS_RE = re.compile(r"^(?P<class_name>[a-z0-9]+)\(.*\)$", re.IGNORECASE)
ARRAY_RE = re.compile(r"array\((?P<contents>\[.*\])\)")

FRAME_KEYS = ("frames", "stim_vsync", "vsync_stim")
PHOTODIODE_KEYS = ("photodiode", "stim_photodiode")
OPTOGENETIC_STIMULATION_KEYS = ("LED_sync", "opto_trial")
EYE_TRACKING_KEYS = (
    "eye_frame_received",  # Expected eye tracking
    # line label after 3/27/2020
    # clocks eye tracking frame pulses (port 0, line 9)
    "cam2_exposure",
    # previous line label for eye tracking
    # (prior to ~ Oct. 2018)
    "eyetracking",
    "eye_cam_exposing",
    "eye_tracking",
)  # An undocumented, but possible eye tracking line label  # NOQA E114
BEHAVIOR_TRACKING_KEYS = (
    "beh_frame_received",  # Expected behavior line label after 3/27/2020  # NOQA E127
    # clocks behavior tracking frame # NOQA E127
    # pulses (port 0, line 8)
    "cam1_exposure",
    "behavior_monitoring",
)


def convert_filepath_caseinsensitive(filename_in):
    """
    Replaces the case of training

    Parameters
    ----------
    filename_in : str
        The filename to convert

    Returns
    -------
    str
        The filename with the case replaced
    """
    return filename_in.replace("TRAINING", "training")


INT_NULL = -1


def enforce_df_int_typing(
    input_df: pd.DataFrame,
    int_columns: List[str],
    use_pandas_type: object = False,
) -> pd.DataFrame:
    """Enforce integer typing for columns that may have lost int typing when
    combined into the final DataFrame.

    Parameters
    ----------
    input_df : pandas.DataFrame
        DataFrame with typing to enforce.
    int_columns : list of str
        Columns to enforce int typing and fill any NaN/None values with
        INT_NULL (-1). Requested columns not in the dataframe are ignored.
    use_pandas_type : bool
        Instead of filling with INT_NULL to enforce integer typing,
        use the pandas type Int64. This type can have issues converting to
        numpy/array type values.

    Returns
    -------
    output_df : pandas.DataFrame
        DataFrame specific columns hard typed to Int64 to allow NA values
        without resorting to float type.
    """
    for col in int_columns:
        if col in input_df.columns:
            if use_pandas_type:
                input_df[col] = input_df[col].astype("Int64")
            else:
                input_df[col] = input_df[col].fillna(INT_NULL).astype(int)
    return input_df


def enforce_df_column_order(input_df: pd.DataFrame, column_order: List[str]) -> pd.DataFrame:
    """Return the data frame but with columns ordered.

    Parameters
    ----------
    input_df : pandas.DataFrame
        Data frame with columns to be ordered.
    column_order : list of str
        Ordering of column names to enforce. Columns not specified are shifted
        to the end of the order but retain their order amongst others not
        specified. If a specified column is not in the DataFrame it is ignored.

    Returns
    -------
    output_df : pandas.DataFrame
        DataFrame the same as the input but with columns reordered.
    """
    # Use only columns that are in the input dataframe's columns.
    pruned_order = []
    for col in column_order:
        if col in input_df.columns:
            pruned_order.append(col)
    # Get the full list of columns in the data frame with our ordered columns
    # first.
    pruned_order.extend(list(set(input_df.columns).difference(set(pruned_order))))
    return input_df[pruned_order]


def seconds_to_frames(seconds, pkl_file):
    """
    Convert seconds to frames using the pkl file.

    Parameters
    ----------
    seconds : list of float
        Seconds to convert to frames.
    pkl_file : str
        Path to the pkl file.

    Returns
    -------
    frames : list of int
        Frames corresponding to the input seconds.
    """
    return (np.array(seconds) + pkl.get_pre_blank_sec(pkl_file)) * pkl.get_fps(pkl_file)


def extract_const_params_from_stim_repr(stim_repr, repr_params_re=REPR_PARAMS_RE, array_re=ARRAY_RE):
    """Parameters which are not set as sweep_params in the stimulus script
    (usually because they are not varied during the course of the session) are
    not output in an easily machine-readable format. This function
    attempts to recover them by parsing the string repr of the stimulus.

    Parameters
    ----------
        stim_repr : str
            The repr of the camstim stimulus object. Served up per-stimulus
            in the stim pickle.
        repr_params_re : re.Pattern
            Extracts attributes as "="-seperated strings
        array_re : re.Pattern
            Extracts list reprs from numpy array reprs.

    Returns
    -------
    repr_params : dict
        dictionary of paramater keys and values extracted from the stim repr.
        Where possible, the values are converted to native Python types.

    """

    repr_params = {}

    for match in repr_params_re.findall(stim_repr):
        k, v = match.split("=")

        if k not in repr_params:
            m = array_re.match(v)
            if m is not None:
                v = m["contents"]

            try:
                v = ast.literal_eval(v)
            except ValueError:
                pass

            repr_params[k] = v

        else:
            raise KeyError(f"duplicate key: {k}")

    return repr_params


def parse_stim_repr(
    stim_repr,
    drop_params=DROP_PARAMS,
    repr_params_re=REPR_PARAMS_RE,
    array_re=ARRAY_RE,
    raise_on_unrecognized=False,
):
    """Read the string representation of a psychopy stimulus and extract
    stimulus parameters.

    Parameters
    ----------
    stim_repr : str
    drop_params : tuple
    repr_params_re : re.Pattern
    array_re : re.Pattern


    Returns
    -------
    dict :
        maps extracted parameter names to values

    """

    stim_params = extract_const_params_from_stim_repr(stim_repr, repr_params_re=repr_params_re, array_re=array_re)

    for drop_param in drop_params:
        if drop_param in stim_params:
            del stim_params[drop_param]

    logger.debug(stim_params)
    return stim_params


def create_stim_table(
    pkl_file,
    stimuli,
    stimulus_tabler,
    spontaneous_activity_tabler,
    sort_key="start_time",
    block_key="stim_block",
    index_key="stim_index",
):
    """Build a full stimulus table

    Parameters
    ----------
    stimuli : list of dict
        Each element is a stimulus dictionary,
        as provided by the stim.pkl file.
    stimulus_tabler : function
        A function which takes a single stimulus dictionary
        as its argument and returns a stimulus table dataframe.
    spontaneous_activity_tabler : function
        A function which takes a list of stimulus tables as
        arguments and returns a list of 0 or more tables
        describing spontaneous activity sweeps.
    sort_key : str, optional
        Sort the final stimulus table in ascending order by this key.
        Defaults to 'start_time'.

    Returns
    -------
    stim_table_full : pandas.DataFrame
        Each row is a sweep. Has columns describing (in frames) the start
        and end times of each sweep. Other columns
        describe the values of stimulus parameters on those sweeps.

    """

    stimulus_tables = []
    for ii, stimulus in enumerate(stimuli):
        current_tables = stimulus_tabler(pkl_file, stimulus)
        for table in current_tables:
            table[index_key] = ii

        stimulus_tables.extend(current_tables)

    stimulus_tables = sorted(stimulus_tables, key=lambda df: min(df[sort_key].values))
    for ii, stim_table in enumerate(stimulus_tables):
        stim_table[block_key] = ii

    stimulus_tables.extend(spontaneous_activity_tabler(stimulus_tables))

    stim_table_full = pd.concat(stimulus_tables, ignore_index=True, sort=False)
    stim_table_full.sort_values(by=[sort_key], inplace=True)
    stim_table_full.reset_index(drop=True, inplace=True)

    return stim_table_full


def make_spontaneous_activity_tables(
    stimulus_tables,
    start_key="start_time",
    end_key="stop_time",
    duration_threshold=0.0,
):
    """Fills in frame gaps in a set of stimulus tables. Suitable for use as
    the spontaneous_activity_tabler in create_stim_table.

    Parameters
    ----------
    stimulus_tables : list of pd.DataFrame
        Input tables - should have start_key and end_key columns.
    start_key : str, optional
        Column name for the start of a sweep. Defaults to 'start_time'.
    end_key : str, optional
        Column name for the end of a sweep. Defaults to 'stop_time'.
    duration_threshold : numeric or None
        If not None (default is 0), remove spontaneous activity sweeps
        whose duration is less than this threshold.

    Returns
    -------
    list :
        Either empty, or contains a single pd.DataFrame.
        The rows of the dataframe are spontaneous activity sweeps.

    """

    nstimuli = len(stimulus_tables)
    if nstimuli == 0:
        return []

    spon_start = np.zeros(nstimuli + 1, dtype=int)
    spon_end = np.zeros(nstimuli, dtype=int)

    for ii, table in enumerate(stimulus_tables):
        spon_start[ii + 1] = table[end_key].values[-1]
        spon_end[ii] = table[start_key].values[0]
        # Assume the same block is being represented twice,
        # so we grab the next relevant block
        if spon_end[ii] < spon_start[ii]:
            temp = spon_end[ii]
            spon_end[ii] = spon_start[ii]
            spon_start[ii] = temp

    spon_start = spon_start[:-1]
    spon_sweeps = pd.DataFrame({start_key: spon_start, end_key: spon_end})

    if duration_threshold is not None:
        spon_sweeps = spon_sweeps[np.fabs(spon_sweeps[start_key] - spon_sweeps[end_key]) > duration_threshold]
        spon_sweeps.reset_index(drop=True, inplace=True)
    spon_sweeps = spon_sweeps.drop_duplicates(subset=[start_key, end_key])
    spon_sweeps.reset_index(drop=True, inplace=True)
    return [spon_sweeps]


def extract_blocks_from_stim(stims):
    """
    Extracts the blocks from a stim with multiple blocks

    Parameters
    ----------
    stims: List[dict]
        A list of stimuli dictionaries

    Returns
    -------
    stim_chunked_blocks: List[dict]
        A list of stimuli dictionaries with the blocks extracted
    """
    stim_chunked_blocks = []
    for stimulus in stims:
        if "stimuli" in stimulus:
            for stimulus_block in stimulus["stimuli"]:
                stim_chunked_blocks.append(stimulus_block)
        else:
            stim_chunked_blocks.append(stimulus)
    return stim_chunked_blocks


def extract_frame_times_from_vsync(
    sync_file,
    frame_keys=FRAME_KEYS,
):
    """
    Extracts frame times from a vsync signal

    Parameters
    ----------
    sync_file : h5py.File
        File containing sync data.
    photodiode_cycle : numeric, optional
        The number of frames between photodiode pulses. Defaults to 60.
    frame_keys : tuple of str, optional
        Keys to extract frame times from. Defaults to FRAME_KEYS.
    photodiode_keys : tuple of str, optional
        Keys to extract photodiode times from. Defaults to PHOTODIODE_KEYS.
    trim_discontiguous_frame_times : bool, optional
        If True, remove discontiguous frame times. Defaults to True.

    Returns
    -------
    frame_start_times : np.ndarray
        The start times of each frame.

    """

    vsync_times = sync.get_edges(sync_file, "falling", frame_keys)
    return vsync_times


def extract_frame_times_from_photodiode(
    sync_file,
    photodiode_cycle=60,
    frame_keys=FRAME_KEYS,
    photodiode_keys=PHOTODIODE_KEYS,
    trim_discontiguous_frame_times=True,
):
    """
    Extracts frame times from a photodiode signal.

    Parameters
    ----------
    sync_file : h5py.File
        File containing sync data.
    photodiode_cycle : numeric, optional
        The number of frames between photodiode pulses. Defaults to 60.
    frame_keys : tuple of str, optional
        Keys to extract frame times from. Defaults to FRAME_KEYS.
    photodiode_keys : tuple of str, optional
        Keys to extract photodiode times from. Defaults to PHOTODIODE_KEYS.
    trim_discontiguous_frame_times : bool, optional
        If True, remove discontiguous frame times. Defaults to True.

    Returns
    -------
    frame_start_times : np.ndarray
        The start times of each frame.

    """

    photodiode_times = sync.get_edges(sync_file, "all", photodiode_keys)
    vsync_times = sync.get_edges(sync_file, "falling", frame_keys)

    if trim_discontiguous_frame_times:
        vsync_times = sync.trim_discontiguous_vsyncs(vsync_times)

    (
        vsync_times_chunked,
        pd_times_chunked,
    ) = sync.separate_vsyncs_and_photodiode_times(vsync_times, photodiode_times, photodiode_cycle)

    frame_start_times = np.zeros((0,))

    for i in range(len(vsync_times_chunked)):
        photodiode_times = sync.trim_border_pulses(pd_times_chunked[i], vsync_times_chunked[i])
        photodiode_times = sync.correct_on_off_effects(photodiode_times)
        photodiode_times = sync.fix_unexpected_edges(photodiode_times, cycle=photodiode_cycle)

        frame_duration = sync.estimate_frame_duration(photodiode_times, cycle=photodiode_cycle)
        irregular_interval_policy = functools.partial(sync.allocate_by_vsync, np.diff(vsync_times_chunked[i]))
        (
            frame_indices,
            frame_starts,
            frame_end_times,
        ) = sync.compute_frame_times(
            photodiode_times,
            frame_duration,
            len(vsync_times_chunked[i]),
            cycle=photodiode_cycle,
            irregular_interval_policy=irregular_interval_policy,
        )

        frame_start_times = np.concatenate((frame_start_times, frame_starts))

    frame_start_times = sync.remove_zero_frames(frame_start_times)

    return frame_start_times


def calculate_frame_mean_time(sync_file, frame_keys):
    """
    Calculate the mean monitor time of the photodiode signa

    Parameters
    ----------
    sync_file : h5py.File
        File containing sync data.
    frame_keys : tuple of str
        Keys to extract frame times from.

    Returns
    -------
    ptd_start : int
        The start index of the photodiode signal.
    ptd_end : int
        The end index of the photodiode signal.
    """
    FIRST_ELEMENT_INDEX = 0
    sample_freq = 100000.0
    ONE = 1
    # photodiode transitions
    photodiode_rise = np.array(
        sync.get_rising_edges(sync_file, "stim_photodiode"),
        dtype=np.float64,
    ) / float(sample_freq)

    # Find start and stop of stimulus
    # test and correct for photodiode transition errors
    photodiode_rise_diff = np.ediff1d(photodiode_rise)
    min_short_photodiode_rise = 0.1
    max_short_photodiode_rise = 0.3
    min_medium_photodiode_rise = 0.5
    max_medium_photodiode_rise = 1.5
    min_large_photodiode_rise = 1.9
    max_large_photodiode_rise = 2.1

    # find the short and medium length photodiode rises
    short_rise_indexes = np.where(
        np.logical_and(
            photodiode_rise_diff > min_short_photodiode_rise,
            photodiode_rise_diff < max_short_photodiode_rise,
        )
    )[FIRST_ELEMENT_INDEX]
    medium_rise_indexes = np.where(
        np.logical_and(
            photodiode_rise_diff > min_medium_photodiode_rise,
            photodiode_rise_diff < max_medium_photodiode_rise,
        )
    )[FIRST_ELEMENT_INDEX]

    short_set = set(short_rise_indexes)

    # iterate through the medium photodiode rise indexes
    # to find the start and stop indexes
    # lookng for three rise pattern
    next_frame = ONE
    start_pattern_index = 2
    end_pattern_index = 3
    ptd_start = None
    ptd_end = None

    # if we can't use medium, we can apply a filter across large
    if len(medium_rise_indexes) < 3:
        large_rise_indexes = np.where(
            np.logical_and(
                photodiode_rise_diff > min_large_photodiode_rise,
                photodiode_rise_diff < max_large_photodiode_rise,
            )
        )[FIRST_ELEMENT_INDEX]
        for large_rise_index in large_rise_indexes:
            if (
                set(
                    range(
                        large_rise_index - start_pattern_index,
                        large_rise_index,
                    )
                )
                <= short_set
            ):
                ptd_start = large_rise_index + next_frame
            elif (
                set(
                    range(
                        large_rise_index + next_frame,
                        large_rise_index + end_pattern_index,
                    )
                )
                <= short_set
            ):
                ptd_end = large_rise_index

    else:
        for medium_rise_index in medium_rise_indexes:
            if (
                set(
                    range(
                        medium_rise_index - start_pattern_index,
                        medium_rise_index,
                    )
                )
                <= short_set
            ):
                ptd_start = medium_rise_index + next_frame
            elif (
                set(
                    range(
                        medium_rise_index + next_frame,
                        medium_rise_index + end_pattern_index,
                    )
                )
                <= short_set
            ):
                ptd_end = medium_rise_index
    return ptd_start, ptd_end


def extract_frame_times_with_delay(
    sync_file,
    frame_keys=FRAME_KEYS,
):
    """
    Extracts frame times from a vsync signal
    Appends monitor delay

    Parameters
    ----------
    sync_file : h5py.File
        File containing sync data.
    frame_keys : tuple of str, optional
        Keys to extract frame times from. Defaults to FRAME_KEYS.

    Returns
    -------
    frame_start_times : np.ndarray
        The start times of each frame.

    """

    ASSUMED_DELAY = 0.0356
    DELAY_THRESHOLD = 0.002
    FIRST_ELEMENT_INDEX = 0
    ROUND_PRECISION = 4
    sample_freq = 100000.0

    stim_vsync_fall = sync.get_edges(sync_file, "falling", frame_keys)

    logger.info("calculating monitor delay")

    # photodiode transitions
    photodiode_rise = np.array(
        sync.get_rising_edges(sync_file, "stim_photodiode"),
        dtype=np.float64,
    ) / float(sample_freq)

    # Find start and stop of stimulus
    # test and correct for photodiode transition errors
    photodiode_rise_diff = np.ediff1d(photodiode_rise)

    try:
        ptd_start, ptd_end = calculate_frame_mean_time(sync_file, frame_keys)
        # if the photodiode signal exists
        if ptd_start is not None and ptd_end is not None:
            # check to make sure there are no there are no photodiode errors
            # sometimes two consecutive photodiode events take place close
            # to each other.
            # Correct this case if it happens
            photodiode_rise_error_threshold = 1.8
            last_frame_index = -1

            # iterate until all of the errors have been corrected
            while any(photodiode_rise_diff[ptd_start:ptd_end] < photodiode_rise_error_threshold):
                error_frames = (
                    np.where(photodiode_rise_diff[ptd_start:ptd_end] < photodiode_rise_error_threshold)[
                        FIRST_ELEMENT_INDEX
                    ]
                    + ptd_start
                )
                # remove the bad photodiode event
                photodiode_rise = np.delete(photodiode_rise, error_frames[last_frame_index])
                ptd_end -= 1

            # Find the delay
            # calculate monitor delay
            first_pulse = ptd_start
            number_of_photodiode_rises = ptd_end - ptd_start
            half_vsync_fall_events_per_photodiode_rise = 60
            vsync_fall_events_per_photodiode_rise = half_vsync_fall_events_per_photodiode_rise * 2

            delay_rise = np.empty(number_of_photodiode_rises)
            for photodiode_rise_index in range(number_of_photodiode_rises):
                delay_rise[photodiode_rise_index] = (
                    photodiode_rise[photodiode_rise_index + first_pulse]
                    - stim_vsync_fall[
                        (photodiode_rise_index * vsync_fall_events_per_photodiode_rise)
                        + half_vsync_fall_events_per_photodiode_rise
                    ]
                )

            # get a single delay value by finding the mean of
            # all of the delays -
            # skip the last element in the array
            delay = np.mean(delay_rise[:last_frame_index])
            delay_std = np.std(delay_rise[:last_frame_index])

            if delay_std > DELAY_THRESHOLD or np.isnan(delay):
                if np.abs((delay - 1) - ASSUMED_DELAY) < DELAY_THRESHOLD:
                    logger.info("One second flip required")
                    return delay - 1

            return delay
        else:
            return ASSUMED_DELAY
    except Exception as e:
        print(e)
        delay = ASSUMED_DELAY
        logger.error("Process without photodiode signal. Assumed delay: {}".format(round(delay, ROUND_PRECISION)))
        return delay


def convert_frames_to_seconds(
    stimulus_table,
    frame_times,
    frames_per_second=None,
    extra_frame_time=False,
    map_columns=("start_time", "stop_time"),
):
    """Converts sweep times from frames to seconds.

    Parameters
    ----------
    stimulus_table : pd.DataFrame
        Rows are sweeps. Columns are stimulus parameters as well as start
        and end frames for each sweep.
    frame_times : numpy.ndarrray
        Gives the time in seconds at which each frame (indices) began.
    frames_per_second : numeric, optional
        If provided, and extra_frame_time is True, will be used to calculcate
        the extra_frame_time.
    extra_frame_time : float, optional
        If provided, an additional frame time will be appended. The time will
        be incremented by extra_frame_time from
        the previous last frame time, to denote the time at which the last
        frame ended. If False, no extra time will be
        appended. If None (default), the increment will be 1.0/fps.
    map_columns : tuple of str, optional
        Which columns to replace with times. Defaults to 'start_time'
        and 'stop_time'

    Returns
    -------
    stimulus_table : pd.DataFrame
        As above, but with map_columns values converted to seconds from frames.

    """

    stimulus_table = stimulus_table.copy()

    if extra_frame_time is True and frames_per_second is not None:
        extra_frame_time = 1.0 / frames_per_second
    if extra_frame_time is not False:
        frame_times = np.append(frame_times, frame_times[-1] + extra_frame_time)

    for column in map_columns:
        stimulus_table[column] = frame_times[np.around(stimulus_table[column]).astype(int)]

    return stimulus_table


def apply_display_sequence(
    sweep_frames_table,
    frame_display_sequence,
    start_key="start_time",
    end_key="stop_time",
    diff_key="dif",
    block_key="stim_block",
):
    """Adjust raw sweep frames for a stimulus based on the display sequence
    for that stimulus.

    Parameters
    ----------
    sweep_frames_table : pd.DataFrame
        Each row is a sweep. Has two columns, 'start' and 'end',
        which describe (in frames) when that sweep began and ended.
    frame_display_sequence : np.ndarray
        2D array. Rows are display intervals. The 0th column is the start
        frame of that interval, the 1st the end frame.

    Returns
    -------
    sweep_frames_table : pd.DataFrame
        As above, but start and end frames have been adjusted based on
        the display sequence.

    Notes
    -----
    The frame values in the raw sweep_frames_table are given in 0-indexed
    offsets from the start of display for this stimulus. This domain only
    takes into account frames which are part of a display interval for that
    stimulus, so the frame ids need to be adjusted to lie on the global
    frame sequence.

    """

    sweep_frames_table = sweep_frames_table.copy()
    if block_key not in sweep_frames_table.columns.values:
        sweep_frames_table[block_key] = np.zeros((sweep_frames_table.shape[0]), dtype=int)

    sweep_frames_table[diff_key] = sweep_frames_table[end_key] - sweep_frames_table[start_key]

    sweep_frames_table[start_key] += frame_display_sequence[0, 0]
    for seg in range(len(frame_display_sequence) - 1):
        match_inds = sweep_frames_table[start_key] >= frame_display_sequence[seg, 1]

        sweep_frames_table.loc[match_inds, start_key] += (
            frame_display_sequence[seg + 1, 0] - frame_display_sequence[seg, 1]
        )
        sweep_frames_table.loc[match_inds, block_key] = seg + 1

    sweep_frames_table[end_key] = sweep_frames_table[start_key] + sweep_frames_table[diff_key]
    sweep_frames_table = sweep_frames_table[sweep_frames_table[end_key] <= frame_display_sequence[-1, 1]]
    sweep_frames_table = sweep_frames_table[sweep_frames_table[start_key] <= frame_display_sequence[-1, 1]]

    sweep_frames_table.drop(diff_key, inplace=True, axis=1)
    return sweep_frames_table


def get_image_set_name(image_set_path: str):
    """
    Strips the stem from the image_set filename
    """
    return Path(image_set_path).stem


def read_stimulus_name_from_path(stimulus):
    """Obtains a human-readable stimulus name by looking at the filename of
    the 'stim_path' item.

    Parameters
    ----------
    stimulus : dict
        must contain a 'stim_path' item.

    Returns
    -------
    str :
        name of stimulus

    """

    if "stim_path" in stimulus:
        stim_name = stimulus["stim_path"]

    if stim_name == "":
        if "movie_local_path" in stimulus and stimulus["movie_local_path"] != "":
            stim_name = stimulus["movie_local_path"].split("\\")[-1].split(".")[0]
        else:
            stim_name = stimulus["stim"]
    else:
        stim_name = stim_name.split("\\")[-1].split(".")[0]
    return stim_name


def get_stimulus_type(stimulus):
    """
    Obtains the stimulus type from the stimulus dictionary.

    Parameters
    ----------
    stimulus : dict
        A dictionary describing a stimulus.

    Returns
    -------
    str :
        The stimulus type.
    """
    input_string = stimulus["stim"]

    # Regex for single quotes
    pattern = r"name='([^']+)'"

    match = re.search(pattern, input_string)

    if match:
        stim_type = match.group(1)
        stim_type = stim_type.replace("unnamed ", "")
        return stim_type
    else:
        return "None or Blank"


def get_stimulus_image_name(stimulus, index):
    """
    Extracts the image name from the stimulus dictionary.
    Used when image is NOT from a movie file.
    """
    image_index = stimulus["sweep_order"][index]
    image_name = stimulus["image_path_list"][image_index]
    # Use regex to capture everything after 'passive\\'
    match = re.search(r"passive\\(.+)", image_name)

    if match:
        extracted_image_name = match.group(1)
    return extracted_image_name


def build_stimuluswise_table(
    pickle_file,
    stimulus,
    seconds_to_frames,
    start_key="start_time",
    end_key="stop_time",
    name_key="stim_name",
    template_key="stim_type",
    block_key="stim_block",
    get_stimulus_name=None,
    extract_const_params_from_repr=False,
    drop_const_params=DROP_PARAMS,
):
    """Construct a table of sweeps, including their times on the
    experiment-global clock and the values of each relevant parameter.

    Parameters
    ----------
    stimulus : dict
        Describes presentation of a stimulus on a particular experiment. Has
        a number of fields, of which we are using:
            stim_path : str
                windows file path to the stimulus data
            sweep_frames : list of lists
                rows are sweeps, columns are start and end frames of that sweep
                (in the stimulus-specific frame domain). C-order.
            sweep_order : list of int
                indices are frames, values are the sweep on that frame
            display_sequence : list of list
                rows are intervals in which the stimulus was displayed.
                Columns are start and end times (s, global) of the display.
                C-order.
             dimnames : list of str
                Names of parameters for this stimulus (such as "Contrast")
            sweep_table : list of tuple
                Each element is a tuple of parameter values (1 per dimname)
                describing a single sweep.
    seconds_to_frames : function
        Converts experiment seconds to frames
    start_key : str, optional
        key to use for start frame indices. Defaults to 'start_time'
    end_key : str, optional
        key to use for end frame indices. Defaults to 'stop_time'
    name_key : str, optional
        key to use for stimulus name annotations. Defaults to 'stim_name'
    block_key : str, optional
        key to use for the 0-index position of this stimulus block
    get_stimulus_name : function | dict -> str, optional
        extracts stimulus name from the stimulus dictionary. Default is
        read_stimulus_name_from_path

    Returns
    -------
    list of pandas.DataFrame :
        Each table corresponds to an entry in the display sequence.
        Rows are sweeps, columns are stimulus parameter values as well as
        "start_time" and 'stop_time.

    """

    if get_stimulus_name is None:
        get_stimulus_name = read_stimulus_name_from_path

    if stimulus["display_sequence"] is None:
        get_stimulus_name = get_stimulus_image_name
        frame_display_sequence = (
            stimulus["sweep_frames"][0][0],
            stimulus["sweep_frames"][-1][1],
        )
        sweep_frames_table = pd.DataFrame(stimulus["sweep_frames"], columns=(start_key, end_key))
        sweep_frames_table[block_key] = np.zeros([sweep_frames_table.shape[0]], dtype=int)
        stim_table = pd.DataFrame(
            {
                start_key: sweep_frames_table[start_key],
                end_key: sweep_frames_table[end_key] + 1,
                name_key: [get_stimulus_name(stimulus, idx) for idx in sweep_frames_table.index],
                template_key: "Image",
                block_key: sweep_frames_table[block_key],
            }
        )
    else:
        frame_display_sequence = seconds_to_frames(stimulus["display_sequence"], pickle_file)
        sweep_frames_table = pd.DataFrame(stimulus["sweep_frames"], columns=(start_key, end_key))
        sweep_frames_table[block_key] = np.zeros([sweep_frames_table.shape[0]], dtype=int)
        sweep_frames_table = apply_display_sequence(sweep_frames_table, frame_display_sequence, block_key=block_key)

        stim_table = pd.DataFrame(
            {
                start_key: sweep_frames_table[start_key],
                end_key: sweep_frames_table[end_key] + 1,
                name_key: get_stimulus_name(stimulus),
                template_key: get_stimulus_type(stimulus),
                block_key: sweep_frames_table[block_key],
            }
        )

    sweep_order = stimulus["sweep_order"][: len(sweep_frames_table)]
    dimnames = stimulus["dimnames"]

    if not dimnames or "ReplaceImage" in dimnames:
        stim_table["Image"] = sweep_order
    else:
        stim_table["sweep_number"] = sweep_order
        sweep_table = pd.DataFrame(stimulus["sweep_table"], columns=dimnames)
        sweep_table["sweep_number"] = sweep_table.index

        stim_table = assign_sweep_values(stim_table, sweep_table)
        stim_table = split_column(
            stim_table,
            "Pos",
            {"Pos_x": lambda field: field[0], "Pos_y": lambda field: field[1]},
        )

    if extract_const_params_from_repr:
        const_params = parse_stim_repr(stimulus["stim"], drop_params=drop_const_params)
        existing_columns = set(stim_table.columns)
        for const_param_key, const_param_value in const_params.items():
            existing_cap = const_param_key.capitalize() in existing_columns
            existing_upper = const_param_key.upper() in existing_columns
            existing = const_param_key in existing_columns

            if not (existing_cap or existing_upper or existing):
                stim_table[const_param_key] = [const_param_value] * stim_table.shape[0]
            else:
                raise KeyError(f"column {const_param_key} already exists")

    unique_indices = np.unique(stim_table[block_key].values)
    output = [stim_table.loc[stim_table[block_key] == ii, :] for ii in unique_indices]

    return output


def split_column(table, column, new_columns, drop_old=True):
    """Divides a dataframe column into multiple columns.

    Parameters
    ----------
    table : pandas.DataFrame
        Columns will be drawn from and assigned to this dataframe. This
        dataframe will NOT be modified inplace.
    column : str
        This column will be split.
    new_columns : dict, mapping strings to functions
        Each key will be the name of a new column, while its value (a function)
        will be used to build the new column's values. The functions should map
        from a single value of the original column to a single value
        of the new column.
    drop_old : bool, optional
        If True, the original column will be dropped from the table.

    Returns
    -------
    table : pd.DataFrame
        The modified table

    """

    if column not in table:
        return table
    table = table.copy()

    for new_column, rule in new_columns.items():
        table[new_column] = table[column].apply(rule)

    if drop_old:
        table.drop(column, inplace=True, axis=1)
    return table


def assign_sweep_values(
    stim_table,
    sweep_table,
    on="sweep_number",
    drop=True,
    tmp_suffix="_stimtable_todrop",
):
    """Left joins a stimulus table to a sweep table in order to associate
        epochs in time with stimulus characteristics.

    Parameters
    ----------
    stim_table : pd.DataFrame
        Each row is a stimulus epoch, with start and end times and a foreign
        key onto a particular sweep.
    sweep_table : pd.DataFrame
        Each row is a sweep. Should have columns in common with the stim_table
        - the resulting table will use values from the sweep_table.
    on : str, optional
        Column on which to join.
    drop : bool, optional
        If True (default), the join column (argument on) will be dropped from
        the output.
    tmp_suffix : str, optional
        Will be used to identify overlapping columns. Should not appear in the
        name of any column in either dataframe.

    """

    joined_table = stim_table.join(sweep_table, on=on, lsuffix=tmp_suffix)
    for dim in joined_table.columns.values:
        if tmp_suffix in dim:
            joined_table.drop(dim, inplace=True, axis=1)

    if drop:
        joined_table.drop(on, inplace=True, axis=1)
    return joined_table



"""Utils to process naming of stimulus columns"""

import logging
import warnings

import numpy as np

import aind_metadata_extractor.utils.camstim_sync.constants as constants

logger = logging.getLogger(__name__)


def drop_empty_columns(table):
    """Remove from the stimulus table columns whose values are all nan"""

    to_drop = []

    for colname in table.columns:
        if table[colname].isna().all():
            to_drop.append(colname)

    table.drop(columns=to_drop, inplace=True)
    return table


def collapse_columns(table):
    """merge, where possible, columns that describe the same parameter. This
    is pretty conservative - it only matches columns by capitalization and
    it only overrides nans.
    """

    colnames = set(table.columns)

    matches = []
    for col in table.columns:
        for transformed in (col.upper(), col.capitalize()):
            if transformed in colnames and col != transformed:
                col_notna = ~(table[col].isna())
                trans_notna = ~(table[transformed].isna())
                if (col_notna & trans_notna).sum() != 0:
                    continue

                mask = ~(col_notna) & (trans_notna)

                matches.append(transformed)
                table.loc[mask, col] = table[transformed][mask]
                break

    table.drop(columns=matches, inplace=True)
    return table


def add_number_to_shuffled_movie(
    table,
    natural_movie_re=constants.GENERIC_MOVIE_RE,
    template_re=constants.SHUFFLED_MOVIE_RE,
    stim_colname="stim_name",
    template="natural_movie_{}_shuffled",
    tmp_colname="__movie_number__",
):
    """
    Adds a number to a shuffled movie stimulus name, if possible.

    Parameters
    ----------
    table : pd.DataFrame
        the incoming stimulus table
    natural_movie_re : re.Pattern, optional
        regex that matches movie stimulus names
    template_re : re.Pattern, optional
        regex that matches shuffled movie stimulus names
    stim_colname : str, optional
        the name of the dataframe column that contains stimulus names
    template : str, optional
        the template's name
    tmp_colname : str, optional
        the name of the template column to use

    Returns
    -------
    table : pd.DataFrame
        the stimulus table with the shuffled movie names updated

    """

    if not table[stim_colname].str.contains(constants.SHUFFLED_MOVIE_RE).any():
        return table
    table = table.copy()

    table[tmp_colname] = table[stim_colname].str.extract(natural_movie_re, expand=True)["number"]

    unique_numbers = [item for item in table[tmp_colname].dropna(inplace=False).unique()]
    if len(unique_numbers) != 1:
        raise ValueError(
            "unable to uniquely determine a movie number for this session. " + f"Candidates: {unique_numbers}"
        )
    movie_number = unique_numbers[0]

    def renamer(row):
        """
        renames the shuffled movie stimulus according to the template

        Parameters
        ----------
        row : pd.Series
            a row in the stimulus table

        Returns
        -------
        table : pd.DataFrame
            the stimulus table with the shuffled movie names updated
        """
        if not isinstance(row[stim_colname], str):
            return row[stim_colname]
        if not template_re.match(row[stim_colname]):
            return row[stim_colname]
        else:
            return template.format(movie_number)

    table[stim_colname] = table.apply(renamer, axis=1)
    logger.debug(table.keys())
    table.drop(columns=tmp_colname, inplace=True)
    return table


def standardize_movie_numbers(
    table,
    movie_re=constants.GENERIC_MOVIE_RE,
    numeral_re=constants.NUMERAL_RE,
    digit_names=constants.DIGIT_NAMES,
    stim_colname="stim_name",
):
    """Natural movie stimuli in visual coding are numbered using words, like
    "natural_movie_two" rather than "natural_movie_2". This function ensures
    that all of the natural movie stimuli in an experiment are named by that
    convention.

    Parameters
    ----------
    table : pd.DataFrame
        the incoming stimulus table
    movie_re : re.Pattern, optional
        regex that matches movie stimulus names
    numeral_re : re.Pattern, optional
        regex that extracts movie numbers from stimulus names
    digit_names : dict, optional
        map from numerals to english words
    stim_colname : str, optional
        the name of the dataframe column that contains stimulus names

    Returns
    -------
    table : pd.DataFrame
        the stimulus table with movie numerals having been mapped to english
        words

    """

    def replace(match_obj):
        """
        replaces the numeral in a movie stimulus name with its english
        equivalent

        Parameters
        ----------
        match_obj : re.Match
            the match object

        Returns
        -------
        str
            the stimulus name with the numeral replaced by its english
            equivalent

        """
        return digit_names[match_obj["number"]]

    # for some reason pandas really wants us to use the captures
    warnings.filterwarnings("ignore", "This pattern has match groups")
    warnings.filterwarnings("ignore", category=UserWarning)

    movie_rows = table[stim_colname].str.contains(movie_re, na=False)
    table.loc[movie_rows, stim_colname] = table.loc[movie_rows, stim_colname].str.replace(
        numeral_re, replace, regex=True
    )

    return table


def map_stimulus_names(table, name_map=None, stim_colname="stim_name"):
    """Applies a mappting to the stimulus names in a stimulus table

    Parameters
    ----------
    table : pd.DataFrame
        the input stimulus table
    name_map : dict, optional
        rename the stimuli according to this mapping
    stim_colname: str, optional
        look in this column for stimulus names

    """

    if name_map is None:
        return table

    name_map[np.nan] = "spontaneous"

    table[stim_colname] = table[stim_colname].replace(to_replace=name_map, inplace=False)

    name_map.pop(np.nan)

    return table


def map_column_names(table, name_map=None, ignore_case=True):
    """
    Maps column names in a table according to a mapping.

    Parameters
    ----------
    table : pd.DataFrame
        the input table
    name_map : dict, optional
        mapping from old names to new names
    ignore_case : bool, optional
        ignore case when mapping column names

    Returns
    -------
    table : pd.DataFrame
        the table with column names mapped

    """
    if ignore_case and name_map is not None:
        name_map = {key.lower(): value for key, value in name_map.items()}

        def mapper(name):
            """
            Maps a column name to a new name from the map

            Parameters
            ----------
            name : str
                the column name to map

            Returns
            -------
            str
                the mapped column name
            """
            name_lower = name.lower()
            if name_lower in name_map:
                return name_map[name_lower]
            return name

    else:
        mapper = name_map

    return table.rename(columns=mapper)
